#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Base functionality for any HTTP-based network component."""
from __future__ import annotations

import abc
import contextlib
import json
import logging
import ssl
import types
import typing

import aiohttp.typedefs

from hikari.net import tracing


class HTTPClient(abc.ABC):  # pylint:disable=too-many-instance-attributes
    """An HTTP client base for Hikari.

    The purpose of this is to provide a consistent interface for any network
    facing components that need an HTTP connection or websocket connection.

    This class takes care of initializing the underlying client session, etc.

    This class will also register interceptors for HTTP requests to produce an
    appropriate level of debugging context when needed.

    Parameters
    ----------
    allow_redirects : bool
        Whether to allow redirects or not. Defaults to `False`.
    connector : aiohttp.BaseConnector, optional
        Optional aiohttp connector info for making an HTTP connection
    debug : bool
        Defaults to `False`. If `True`, then a lot of contextual information
        regarding low-level HTTP communication will be logged to the debug
        logger on this class.
    json_deserialize : deserialization function
        A custom JSON deserializer function to use. Defaults to `json.loads`.
    json_serialize : serialization function
        A custom JSON serializer function to use. Defaults to `json.dumps`.
    proxy_headers : typing.Mapping[str, str], optional
        Optional proxy headers to pass to HTTP requests.
    proxy_auth : aiohttp.BasicAuth, optional
        Optional authorization to be used if using a proxy.
    proxy_url : str, optional
        Optional proxy URL to use for HTTP requests.
    ssl_context : ssl.SSLContext, optional
        The optional SSL context to be used.
    verify_ssl : bool
        Whether or not the client should enforce SSL signed certificate
        verification. If 1 it will ignore potentially malicious
        SSL certificates.
    timeout : float, optional
        The optional timeout for all HTTP requests.
    trust_env : bool
        If `True`, and no proxy info is given, then `HTTP_PROXY` and
        `HTTPS_PROXY` will be used from the environment variables if present.
        Any proxy credentials will be read from the user's `netrc` file
        (https://www.gnu.org/software/inetutils/manual/html_node/The-_002enetrc-file.html)
        If `False`, then this information is instead ignored.
        Defaults to `False`.
    """

    __slots__ = (
        "__client_session",
        "allow_redirects",
        "connector",
        "debug",
        "logger",
        "json_deserialize",
        "json_serialize",
        "proxy_auth",
        "proxy_headers",
        "proxy_url",
        "ssl_context",
        "timeout",
        "tracers",
        "trust_env",
        "verify_ssl",
    )

    GET: typing.Final[str] = "get"
    POST: typing.Final[str] = "post"
    PATCH: typing.Final[str] = "patch"
    PUT: typing.Final[str] = "put"
    HEAD: typing.Final[str] = "head"
    DELETE: typing.Final[str] = "delete"
    OPTIONS: typing.Final[str] = "options"
    APPLICATION_JSON: typing.Final[str] = "application/json"
    APPLICATION_X_WWW_FORM_URLENCODED: typing.Final[str] = "application/x-www-form-urlencoded"
    APPLICATION_OCTET_STREAM: typing.Final[str] = "application/octet-stream"

    allow_redirects: bool
    """`True` if HTTP redirects are enabled, or `False` otherwise."""

    connector: typing.Optional[aiohttp.BaseConnector]
    """The base connector for the `aiohttp.ClientSession`, if provided."""

    debug: bool
    """`True` if debug mode is enabled. `False` otherwise."""

    logger: logging.Logger
    """The logger to use for this object."""

    json_deserialize: typing.Callable[[typing.AnyStr], typing.Any]
    """The JSON deserialization function.

    This consumes a JSON string and produces some object.
    """

    json_serialize: typing.Callable[[typing.Any], typing.AnyStr]
    """The JSON deserialization function.

    This consumes an object and produces some JSON string.
    """

    proxy_auth: typing.Optional[aiohttp.BasicAuth]
    """Proxy authorization to use."""

    proxy_headers: typing.Optional[typing.Mapping[str, str]]
    """A set of headers to provide to a proxy server."""

    proxy_url: typing.Optional[str]
    """An optional proxy URL to send requests to."""

    ssl_context: typing.Optional[ssl.SSLContext]
    """The custom SSL context to use."""

    timeout: typing.Optional[float]
    """The HTTP request timeout to abort requests after."""

    tracers: typing.List[tracing.BaseTracer]
    """Request tracers.

    These can be used to intercept HTTP request events on a low level.
    """

    trust_env: bool
    """Whether to take notice of proxy environment variables.

    If `True`, and no proxy info is given, then `HTTP_PROXY` and
    `HTTPS_PROXY` will be used from the environment variables if present.
    Any proxy credentials will be read from the user's `netrc` file
    (https://www.gnu.org/software/inetutils/manual/html_node/The-_002enetrc-file.html)
    If `False`, then this information is instead ignored.
    """

    verify_ssl: bool
    """Whether SSL certificates should be verified for each request.

    When this is `True` then an exception will be raised whenever invalid SSL
    certificates are received. When this is `False` unrecognised certificates
    that may be illegitimate are accepted and ignored.
    """

    def __init__(
        self,
        *,
        allow_redirects: bool = False,
        connector: typing.Optional[aiohttp.BaseConnector] = None,
        debug: bool = False,
        json_deserialize: typing.Callable[[typing.AnyStr], typing.Dict] = json.loads,
        json_serialize: typing.Callable[[typing.Dict], typing.AnyStr] = json.dumps,
        logger_name: typing.Optional[str] = None,
        proxy_auth: typing.Optional[aiohttp.BasicAuth] = None,
        proxy_headers: typing.Optional[aiohttp.typedefs.LooseHeaders] = None,
        proxy_url: typing.Optional[str] = None,
        ssl_context: typing.Optional[ssl.SSLContext] = None,
        verify_ssl: bool = True,
        timeout: typing.Optional[float] = None,
        trust_env: bool = False,
    ) -> None:
        self.logger = logging.getLogger(
            f"{type(self).__module__}.{type(self).__qualname__}" if logger_name is None else logger_name
        )

        self.__client_session = None
        self.allow_redirects = allow_redirects
        self.connector = connector
        self.debug = debug
        self.json_serialize = json_serialize
        self.json_deserialize = json_deserialize
        self.proxy_auth = proxy_auth
        self.proxy_headers = proxy_headers
        self.proxy_url = proxy_url
        self.ssl_context: ssl.SSLContext = ssl_context
        self.timeout = timeout
        self.trust_env = trust_env
        self.tracers = [(tracing.DebugTracer(self.logger) if debug else tracing.CFRayTracer(self.logger))]
        self.verify_ssl = verify_ssl

    async def __aenter__(self) -> HTTPClient:
        return self

    async def __aexit__(
        self, exc_type: typing.Type[BaseException], exc_val: BaseException, exc_tb: types.TracebackType
    ) -> None:
        await self.close()

    async def close(self) -> None:
        """Close the client safely."""
        with contextlib.suppress(Exception):
            await self.__client_session.close()
            self.__client_session = None
            self.logger.debug("closed %s", type(self).__qualname__)

    def _acquire_client_session(self) -> aiohttp.ClientSession:
        """Acquire a client session to make requests with.

        Returns
        -------
        aiohttp.ClientSession
            The client session to use for requests.

        !!! warn
            This must only be accessed within an asyncio event loop, otherwise
            there is a risk that the session will not have the correct event
            loop; hence why this is private.
        """
        if self.__client_session is None:
            self.__client_session = aiohttp.ClientSession(
                connector=self.connector,
                trust_env=self.trust_env,
                version=aiohttp.HttpVersion11,
                json_serialize=self.json_serialize or json.dumps,
                trace_configs=[t.trace_config for t in self.tracers],
            )
        return self.__client_session

    async def _perform_request(
        self,
        *,
        method: str,
        url: str,
        headers: aiohttp.typedefs.LooseHeaders,
        body: typing.Union[aiohttp.FormData, dict, list, None],
        query: typing.Dict[str, str],
    ) -> aiohttp.ClientResponse:
        """Make an HTTP request and return the response.

        Parameters
        ----------
        method : str
            The verb to use.
        url : str
            The URL to hit.
        headers : typing.Dict[str, str]
            Headers to use when making the request.
        body : typing.Union[aiohttp.FormData, dict, list, None]
            The body to send. Currently this will send the content in
            a form body if you pass an instance of `aiohttp.FormData`, or
            as a JSON body if you pass a `list` or `dict`. Any other types
            will be ignored.
        query : typing.Dict[str, str]
            Mapping of query string arguments to pass in the URL.

        Returns
        -------
        aiohttp.ClientResponse
            The HTTP response.
        """
        if isinstance(body, (dict, list)):
            body = bytes(self.json_serialize(body), "utf-8")
            headers["content-type"] = self.APPLICATION_JSON

        trace_request_ctx = types.SimpleNamespace()
        trace_request_ctx.request_body = body

        return await self._acquire_client_session().request(
            method=method,
            url=url,
            params=query,
            headers=headers,
            data=body,
            allow_redirects=self.allow_redirects,
            proxy=self.proxy_url,
            proxy_auth=self.proxy_auth,
            proxy_headers=self.proxy_headers,
            verify_ssl=self.verify_ssl,
            ssl_context=self.ssl_context,
            timeout=self.timeout,
            trace_request_ctx=trace_request_ctx,
        )

    async def _create_ws(
        self, url: str, *, compress: int = 0, autoping: bool = True, max_msg_size: int = 0
    ) -> aiohttp.ClientWebSocketResponse:
        """Create a websocket.

        Parameters
        ----------
        url : str
            The URL to connect the websocket to.
        compress : int
            The compression type to use, as an int value. Use `0` to disable
            compression.
        autoping : bool
            If `True`, the client will manage automatically pinging/ponging
            in the background. If `False`, this will not occur.
        max_msg_size : int
            The maximum message size to allow to be received. If `0`, then
            no max limit is set.

        Returns
        -------
        aiohttp.ClientWebsocketResponse
            The websocket to use.
        """
        self.logger.debug("creating underlying websocket object from HTTP session")
        return await self._acquire_client_session().ws_connect(
            url=url,
            compress=compress,
            autoping=autoping,
            max_msg_size=max_msg_size,
            proxy=self.proxy_url,
            proxy_auth=self.proxy_auth,
            proxy_headers=self.proxy_headers,
            verify_ssl=self.verify_ssl,
            ssl_context=self.ssl_context,
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""A bot client might go here... eventually..."""

from __future__ import annotations

__all__ = ["BotBase"]

import abc
import asyncio
import datetime
import inspect
import logging
import math
import os
import platform
import time
import typing

from hikari import _about
from hikari.clients import components as _components
from hikari.clients import configs
from hikari.clients import runnable
from hikari.clients import shard_states
from hikari.events import other
from hikari.internal import assertions
from hikari.internal import conversions
from hikari.state import dispatchers

if typing.TYPE_CHECKING:
    from hikari import gateway_entities
    from hikari import guilds
    from hikari import intents
    from hikari.clients import rest as _rest
    from hikari.clients import shards as _shards
    from hikari.events import base as event_base
    from hikari.internal import more_typing
    from hikari.state import event_managers


class BotBase(
    _components.Components, runnable.RunnableClient, dispatchers.EventDispatcher, abc.ABC,
):
    """An abstract base class for a bot implementation.

    Parameters
    ----------
    config : hikari.clients.configs.BotConfig
        The config object to use.
    **kwargs
        Parameters to use to create a hikari.clients.configs.BotConfig from,
        instead of passing a raw config object.

    Examples
    --------
        # You can use it like this:
        config = hikari.BotConfig(token="...", ...)
        bot = hikari.StatelessBot(config=config)

        # Or like this:
        bot = hikari.StatelessBot(token="...", ...)

    """

    logger: logging.Logger
    """The logger to use for this bot."""

    def __init__(self, *, config: typing.Optional[configs.BotConfig] = None, **kwargs: typing.Any) -> None:
        assertions.assert_that(
            bool(config) ^ bool(kwargs), "You must only specify a config object OR kwargs;", TypeError,
        )

        runnable.RunnableClient.__init__(self, logging.getLogger(f"hikari.{type(self).__qualname__}"))

        # noinspection PyArgumentList
        _components.Components.__init__(
            self, config=None, event_dispatcher=None, event_manager=None, rest=None, shards={},
        )

        self._is_shutting_down = False
        self.config = configs.BotConfig(**kwargs) if config is None else config
        self.event_dispatcher = self._create_event_dispatcher(self.config)
        self.event_manager = self._create_event_manager(self)
        self.rest = self._create_rest(self)

    @property
    def heartbeat_latency(self) -> float:
        """Average heartbeat latency for all valid shards.

        This will return a mean of all the heartbeat intervals for all shards
        with a valid heartbeat latency that are in the
        `hikari.clients.shard_states.ShardState.READY` state.

        If no shards are in this state, this will return `float("nan")`
        instead.

        Returns
        -------
        float
            The mean latency for all `READY` shards that have sent at least
            one acknowledged `HEARTBEAT` payload. If there is not at least
            one shard that meets this criteria, this will instead return
            `float("nan")`.
        """
        latencies = []
        for shard in self.shards.values():
            if not math.isnan(shard.heartbeat_latency):
                latencies.append(shard.heartbeat_latency)

        return sum(latencies) / len(latencies) if latencies else float("nan")

    @property
    def total_disconnect_count(self) -> int:
        """Total number of times any shard has disconnected."""
        return sum(s.disconnect_count for s in self.shards.values())

    @property
    def total_reconnect_count(self) -> int:
        """Total number of times any shard has reconnected."""
        return sum(s.reconnect_count for s in self.shards.values())

    @property
    def intents(self) -> typing.Optional[intents.Intent]:  # noqa: D401
        """Intents that are in use for the connection.

        If intents are not being used at all, then this will be `None` instead.
        """
        return self.config.intents

    @property
    def version(self) -> float:
        """Version being used for the gateway API."""
        return self.config.gateway_version

    # noinspection PyTypeChecker,PyUnresolvedReferences
    async def start(self) -> None:
        """Start the bot.

        This will query Discord for the optimal number of shards to use if
        you did not provide an explicit sharding configuration.

        Each required shard is then started up incrementally at a rate that
        reduces the chance of `INVALID_SESSION` spam occurring. After each
        shard websocket has fired the `READY` event, this coroutine will return.

        After invoking this coroutine, you should keep the application alive
        by awaiting the `join` coroutine in this class.
        """
        if self.shards:  # pylint: disable=access-member-before-definition
            raise RuntimeError("Bot is already running.")
        self._is_shutting_down = False

        version = _about.__version__
        path = os.path.abspath(os.path.dirname(inspect.getsourcefile(_about)))
        py_impl = platform.python_implementation()
        py_ver = platform.python_version()
        py_compiler = platform.python_compiler()
        self.logger.info(
            "hikari v%s (installed in %s) (%s %s %s)", version, path, py_impl, py_ver, py_compiler,
        )

        await self._calculate_shards()

        self.logger.info("starting %s", conversions.pluralize(len(self.shards), "shard"))

        start_time = time.perf_counter()

        for i, shard_id in enumerate(self.shards):
            if i > 0:
                self.logger.info("idling for 5 seconds to avoid an invalid session")
                await asyncio.sleep(5)

            shard_obj = self.shards[shard_id]
            await shard_obj.start()

        finish_time = time.perf_counter()

        self.logger.info("started %s shard(s) in approx %.2fs", len(self.shards), finish_time - start_time)

        if self.event_manager is not None:
            await self.dispatch_event(other.StartedEvent())

    async def _calculate_shards(self):
        gateway_bot = await self.rest.fetch_gateway_bot()

        self.logger.info(
            "you have sent an IDENTIFY %s time(s) before now, and have %s remaining. This will reset at %s.",
            gateway_bot.session_start_limit.total - gateway_bot.session_start_limit.remaining,
            gateway_bot.session_start_limit.remaining,
            datetime.datetime.now() + gateway_bot.session_start_limit.reset_after,
        )

        shard_count = self.config.shard_count if self.config.shard_count else gateway_bot.shard_count
        shard_ids = self.config.shard_ids if self.config.shard_ids else range(shard_count)
        url = gateway_bot.url

        self.logger.info("will connect shards to %s", url)

        shard_clients = {}
        for shard_id in shard_ids:
            shard = self._create_shard(self, shard_id, shard_count, url)
            shard_clients[shard_id] = shard

        self.shards = shard_clients  # pylint: disable=attribute-defined-outside-init

    async def join(self) -> None:
        """Wait for each shard to terminate, then return."""
        await asyncio.gather(*(shard_obj.join() for shard_obj in self.shards.values()))

    async def close(self) -> None:
        try:
            if self.shards and not self._is_shutting_down:
                self._is_shutting_down = True
                self.logger.info("stopping %s shard(s)", len(self.shards))
                start_time = time.perf_counter()
                try:
                    await self.dispatch_event(other.StoppingEvent())
                    await asyncio.gather(*(shard_obj.close() for shard_obj in self.shards.values()))
                finally:
                    finish_time = time.perf_counter()
                    self.logger.info("stopped %s shard(s) in approx %.2fs", len(self.shards), finish_time - start_time)
                    await self.dispatch_event(other.StoppedEvent())
        finally:
            await self.rest.close()

    def add_listener(
        self, event_type: typing.Type[dispatchers.EventT], callback: dispatchers.EventCallbackT, **kwargs
    ) -> dispatchers.EventCallbackT:
        return self.event_dispatcher.add_listener(event_type, callback, _stack_level=4)

    def remove_listener(
        self, event_type: typing.Type[dispatchers.EventT], callback: dispatchers.EventCallbackT
    ) -> dispatchers.EventCallbackT:
        return self.event_dispatcher.remove_listener(event_type, callback)

    def wait_for(
        self,
        event_type: typing.Type[dispatchers.EventT],
        *,
        timeout: typing.Optional[float],
        predicate: dispatchers.PredicateT,
    ) -> more_typing.Future:
        return self.event_dispatcher.wait_for(event_type, timeout=timeout, predicate=predicate)

    def dispatch_event(self, event: event_base.HikariEvent) -> more_typing.Future[typing.Any]:
        return self.event_dispatcher.dispatch_event(event)

    async def update_presence(
        self,
        *,
        status: guilds.PresenceStatus = ...,
        activity: typing.Optional[gateway_entities.Activity] = ...,
        idle_since: typing.Optional[datetime.datetime] = ...,
        is_afk: bool = ...,
    ) -> None:
        """Update the presence of the user for all shards.

        This will only update arguments that you explicitly specify a value for.
        Any arguments that you do not explicitly provide some value for will
        not be changed.

        !!! warning
            This will only apply to connected shards.

        !!! note
            If you wish to update a presence for a specific shard, you can do this
            by using the `shards` `typing.Mapping` to find the shard you wish to
            update.

        Parameters
        ----------
        status : hikari.guilds.PresenceStatus
            If specified, the new status to set.
        activity : hikari.gateway_entities.Activity, optional
            If specified, the new activity to set.
        idle_since : datetime.datetime, optional
            If specified, the time to show up as being idle since,
            or `None` if not applicable.
        is_afk : bool
            If specified, `True` if the user should be marked as AFK,
            or `False` otherwise.
        """
        await asyncio.gather(
            *(
                s.update_presence(status=status, activity=activity, idle_since=idle_since, is_afk=is_afk)
                for s in self.shards.values()
                if s.connection_state in (shard_states.ShardState.WAITING_FOR_READY, shard_states.ShardState.READY)
            )
        )

    @staticmethod
    @abc.abstractmethod
    def _create_shard(
        components: _components.Components, shard_id: int, shard_count: int, url: str
    ) -> _shards.ShardClient:
        """Return a new shard for the given parameters.

        Parameters
        ----------
        components : hikari.clients.components.Components
            The components to register.
        shard_id : int
            The shard ID to use.
        shard_count : int
            The shard count to use.
        url : str
            The gateway URL to connect to.

        Returns
        -------
        hikari.clients.shards.ShardClient
            The shard client implementation to use for the given shard ID.

        !!! note
            The `shard_id` and `shard_count` may be set within the `config`
            object passed, but any conforming implementations are expected to
            use the value passed in the `shard_id` and `shard_count` parameters
            regardless. Failure to do so may result in an invalid sharding
            configuration being used.
        """

    @staticmethod
    @abc.abstractmethod
    def _create_rest(components: _components.Components) -> _rest.RESTClient:
        """Return a new REST client from the given configuration.

        Parameters
        ----------
        components : hikari.clients.components.Components
            The components to register.

        Returns
        -------
        hikari.clients.rest.RESTClient
            The REST client to use.
        """

    @staticmethod
    @abc.abstractmethod
    def _create_event_manager(components: _components.Components) -> event_managers.EventManager:
        """Return a new instance of an event manager implementation.

        Parameters
        ----------
        components : hikari.clients.components.Components
            The components to register.

        Returns
        -------
        hikari.state.event_managers.EventManager
            The event manager to use internally.
        """

    @staticmethod
    @abc.abstractmethod
    def _create_event_dispatcher(config: configs.BotConfig) -> dispatchers.EventDispatcher:
        """Return a new instance of an event dispatcher implementation.

        Parameters
        ----------
        config : hikari.clients.configs.BotConfig
            The bot config to use.

        Returns
        -------
        hikari.state.dispatchers.EventDispatcher
        """



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Components and entities that are used to describe Discord gateway message events."""

from __future__ import annotations

__all__ = [
    "MessageCreateEvent",
    "MessageUpdateEvent",
    "MessageDeleteEvent",
    "MessageDeleteBulkEvent",
    "MessageReactionAddEvent",
    "MessageReactionRemoveEvent",
    "MessageReactionRemoveAllEvent",
    "MessageReactionRemoveEmojiEvent",
]

import typing

import attr

from hikari import applications
from hikari import bases as base_entities
from hikari import embeds as _embeds
from hikari import emojis
from hikari import guilds
from hikari import intents
from hikari import messages
from hikari import unset
from hikari import users
from hikari.events import base as base_events
from hikari.internal import conversions
from hikari.internal import marshaller

if typing.TYPE_CHECKING:
    import datetime

    from hikari.internal import more_typing


@base_events.requires_intents(intents.Intent.GUILD_MESSAGES, intents.Intent.DIRECT_MESSAGES)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class MessageCreateEvent(base_events.HikariEvent, messages.Message):
    """Used to represent Message Create gateway events."""


def _deserialize_object_mentions(payload: more_typing.JSONArray) -> typing.Set[base_entities.Snowflake]:
    return {base_entities.Snowflake(mention["id"]) for mention in payload}


def _deserialize_mentions(payload: more_typing.JSONArray) -> typing.Set[base_entities.Snowflake]:
    return {base_entities.Snowflake(mention) for mention in payload}


def _deserialize_attachments(
    payload: more_typing.JSONArray, **kwargs: typing.Any
) -> typing.Sequence[messages.Attachment]:
    return [messages.Attachment.deserialize(attachment, **kwargs) for attachment in payload]


def _deserialize_embeds(payload: more_typing.JSONArray, **kwargs: typing.Any) -> typing.Sequence[_embeds.Embed]:
    return [_embeds.Embed.deserialize(embed, **kwargs) for embed in payload]


def _deserialize_reaction(payload: more_typing.JSONArray, **kwargs: typing.Any) -> typing.Sequence[messages.Reaction]:
    return [messages.Reaction.deserialize(reaction, **kwargs) for reaction in payload]


# This is an arbitrarily partial version of `messages.Message`
@base_events.requires_intents(intents.Intent.GUILD_MESSAGES, intents.Intent.DIRECT_MESSAGES)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class MessageUpdateEvent(base_events.HikariEvent, base_entities.UniqueEntity, marshaller.Deserializable):
    """Represents Message Update gateway events.

    !!! note
        All fields on this model except `MessageUpdateEvent.channel_id` and
        `MessageUpdateEvent.id` may be set to `hikari.unset.UNSET` (a singleton)
        we have not received information about their state from Discord
        alongside field nullability.
    """

    # FIXME: the id here is called "id", but in MessageDeleteEvent it is "message_id"...

    channel_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake, repr=True)
    """The ID of the channel that the message was sent in."""

    guild_id: typing.Union[base_entities.Snowflake, unset.Unset] = marshaller.attrib(
        deserializer=base_entities.Snowflake, if_undefined=unset.Unset, default=unset.UNSET, repr=True
    )
    """The ID of the guild that the message was sent in."""

    author: typing.Union[users.User, unset.Unset] = marshaller.attrib(
        deserializer=users.User.deserialize, if_undefined=unset.Unset, default=unset.UNSET, repr=True
    )
    """The author of this message."""

    # TODO: can we merge member and author together?
    # We could override deserialize to to this and then reorganise the payload, perhaps?
    member: typing.Union[guilds.GuildMember, unset.Unset] = marshaller.attrib(
        deserializer=guilds.GuildMember.deserialize, if_undefined=unset.Unset, default=unset.UNSET
    )
    """The member properties for the message's author."""

    content: typing.Union[str, unset.Unset] = marshaller.attrib(
        deserializer=str, if_undefined=unset.Unset, default=unset.UNSET
    )
    """The content of the message."""

    timestamp: typing.Union[datetime.datetime, unset.Unset] = marshaller.attrib(
        deserializer=conversions.parse_iso_8601_ts, if_undefined=unset.Unset, default=unset.UNSET
    )
    """The timestamp that the message was sent at."""

    edited_timestamp: typing.Union[datetime.datetime, unset.Unset, None] = marshaller.attrib(
        deserializer=conversions.parse_iso_8601_ts, if_none=None, if_undefined=unset.Unset, default=unset.UNSET
    )
    """The timestamp that the message was last edited at.

    Will be `None` if the message wasn't ever edited.
    """

    is_tts: typing.Union[bool, unset.Unset] = marshaller.attrib(
        raw_name="tts", deserializer=bool, if_undefined=unset.Unset, default=unset.UNSET
    )
    """Whether the message is a TTS message."""

    is_mentioning_everyone: typing.Union[bool, unset.Unset] = marshaller.attrib(
        raw_name="mention_everyone", deserializer=bool, if_undefined=unset.Unset, default=unset.UNSET
    )
    """Whether the message mentions `@everyone` or `@here`."""

    user_mentions: typing.Union[typing.Set[base_entities.Snowflake], unset.Unset] = marshaller.attrib(
        raw_name="mentions", deserializer=_deserialize_object_mentions, if_undefined=unset.Unset, default=unset.UNSET,
    )
    """The users the message mentions."""

    role_mentions: typing.Union[typing.Set[base_entities.Snowflake], unset.Unset] = marshaller.attrib(
        raw_name="mention_roles", deserializer=_deserialize_mentions, if_undefined=unset.Unset, default=unset.UNSET,
    )
    """The roles the message mentions."""

    channel_mentions: typing.Union[typing.Set[base_entities.Snowflake], unset.Unset] = marshaller.attrib(
        raw_name="mention_channels",
        deserializer=_deserialize_object_mentions,
        if_undefined=unset.Unset,
        default=unset.UNSET,
    )
    """The channels the message mentions."""

    attachments: typing.Union[typing.Sequence[messages.Attachment], unset.Unset] = marshaller.attrib(
        deserializer=_deserialize_attachments, if_undefined=unset.Unset, default=unset.UNSET, inherit_kwargs=True,
    )
    """The message attachments."""

    embeds: typing.Union[typing.Sequence[_embeds.Embed], unset.Unset] = marshaller.attrib(
        deserializer=_deserialize_embeds, if_undefined=unset.Unset, default=unset.UNSET, inherit_kwargs=True,
    )
    """The message's embeds."""

    reactions: typing.Union[typing.Sequence[messages.Reaction], unset.Unset] = marshaller.attrib(
        deserializer=_deserialize_reaction, if_undefined=unset.Unset, default=unset.UNSET, inherit_kwargs=True
    )
    """The message's reactions."""

    is_pinned: typing.Union[bool, unset.Unset] = marshaller.attrib(
        raw_name="pinned", deserializer=bool, if_undefined=unset.Unset, default=unset.UNSET
    )
    """Whether the message is pinned."""

    webhook_id: typing.Union[base_entities.Snowflake, unset.Unset] = marshaller.attrib(
        deserializer=base_entities.Snowflake, if_undefined=unset.Unset, default=unset.UNSET
    )
    """If the message was generated by a webhook, the webhook's ID."""

    type: typing.Union[messages.MessageType, unset.Unset] = marshaller.attrib(
        deserializer=messages.MessageType, if_undefined=unset.Unset, default=unset.UNSET
    )
    """The message's type."""

    activity: typing.Union[messages.MessageActivity, unset.Unset] = marshaller.attrib(
        deserializer=messages.MessageActivity.deserialize,
        if_undefined=unset.Unset,
        default=unset.UNSET,
        inherit_kwargs=True,
    )
    """The message's activity."""

    application: typing.Optional[applications.Application] = marshaller.attrib(
        deserializer=applications.Application.deserialize,
        if_undefined=unset.Unset,
        default=unset.UNSET,
        inherit_kwargs=True,
    )
    """The message's application."""

    message_reference: typing.Union[messages.MessageCrosspost, unset.Unset] = marshaller.attrib(
        deserializer=messages.MessageCrosspost.deserialize,
        if_undefined=unset.Unset,
        default=unset.UNSET,
        inherit_kwargs=True,
    )
    """The message's cross-posted reference data."""

    flags: typing.Union[messages.MessageFlag, unset.Unset] = marshaller.attrib(
        deserializer=messages.MessageFlag, if_undefined=unset.Unset, default=unset.UNSET
    )
    """The message's flags."""

    nonce: typing.Union[str, unset.Unset] = marshaller.attrib(
        deserializer=str, if_undefined=unset.Unset, default=unset.UNSET
    )
    """The message nonce.

    This is a string used for validating a message was sent.
    """


@base_events.requires_intents(intents.Intent.GUILD_MESSAGES, intents.Intent.DIRECT_MESSAGES)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class MessageDeleteEvent(base_events.HikariEvent, marshaller.Deserializable):
    """Used to represent Message Delete gateway events.

    Sent when a message is deleted in a channel we have access to.
    """

    # TODO: common base class for Message events.

    channel_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake, repr=True)
    """The ID of the channel where this message was deleted."""

    guild_id: typing.Optional[base_entities.Snowflake] = marshaller.attrib(
        deserializer=base_entities.Snowflake, if_undefined=None, default=None, repr=True
    )
    """The ID of the guild where this message was deleted.

    This will be `None` if this message was deleted in a DM channel.
    """

    message_id: base_entities.Snowflake = marshaller.attrib(
        raw_name="id", deserializer=base_entities.Snowflake, repr=True
    )
    """The ID of the message that was deleted."""


def _deserialize_message_ids(payload: more_typing.JSONArray) -> typing.Set[base_entities.Snowflake]:
    return {base_entities.Snowflake(message_id) for message_id in payload}


@base_events.requires_intents(intents.Intent.GUILD_MESSAGES)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class MessageDeleteBulkEvent(base_events.HikariEvent, marshaller.Deserializable):
    """Used to represent Message Bulk Delete gateway events.

    Sent when multiple messages are deleted in a channel at once.
    """

    channel_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake, repr=True)
    """The ID of the channel these messages have been deleted in."""

    guild_id: typing.Optional[base_entities.Snowflake] = marshaller.attrib(
        deserializer=base_entities.Snowflake, if_none=None, repr=True,
    )
    """The ID of the channel these messages have been deleted in.

    This will be `None` if these messages were bulk deleted in a DM channel.
    """

    message_ids: typing.Set[base_entities.Snowflake] = marshaller.attrib(
        raw_name="ids", deserializer=_deserialize_message_ids
    )
    """A collection of the IDs of the messages that were deleted."""


@base_events.requires_intents(intents.Intent.GUILD_MESSAGE_REACTIONS, intents.Intent.DIRECT_MESSAGE_REACTIONS)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class MessageReactionAddEvent(base_events.HikariEvent, marshaller.Deserializable):
    """Used to represent Message Reaction Add gateway events."""

    # TODO: common base classes!

    user_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake, repr=True)
    """The ID of the user adding the reaction."""

    channel_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake, repr=True)
    """The ID of the channel where this reaction is being added."""

    message_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake, repr=True)
    """The ID of the message this reaction is being added to."""

    guild_id: typing.Optional[base_entities.Snowflake] = marshaller.attrib(
        deserializer=base_entities.Snowflake, if_undefined=None, default=None, repr=True
    )
    """The ID of the guild where this reaction is being added.

    This will be `None` if this is happening in a DM channel.
    """

    # TODO: does this contain a user? If not, should it be a PartialGuildMember?
    member: typing.Optional[guilds.GuildMember] = marshaller.attrib(
        deserializer=guilds.GuildMember.deserialize, if_undefined=None, default=None, inherit_kwargs=True
    )
    """The member object of the user who's adding this reaction.

    This will be `None` if this is happening in a DM channel.
    """

    emoji: typing.Union[emojis.UnknownEmoji, emojis.UnicodeEmoji] = marshaller.attrib(
        deserializer=emojis.deserialize_reaction_emoji, inherit_kwargs=True, repr=True
    )
    """The object of the emoji being added."""


@base_events.requires_intents(intents.Intent.GUILD_MESSAGE_REACTIONS, intents.Intent.DIRECT_MESSAGE_REACTIONS)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class MessageReactionRemoveEvent(base_events.HikariEvent, marshaller.Deserializable):
    """Used to represent Message Reaction Remove gateway events."""

    user_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake, repr=True)
    """The ID of the user who is removing their reaction."""

    channel_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake, repr=True)
    """The ID of the channel where this reaction is being removed."""

    message_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake, repr=True)
    """The ID of the message this reaction is being removed from."""

    guild_id: typing.Optional[base_entities.Snowflake] = marshaller.attrib(
        deserializer=base_entities.Snowflake, if_undefined=None, default=None, repr=True
    )
    """The ID of the guild where this reaction is being removed

    This will be `None` if this event is happening in a DM channel.
    """

    emoji: typing.Union[emojis.UnicodeEmoji, emojis.UnknownEmoji] = marshaller.attrib(
        deserializer=emojis.deserialize_reaction_emoji, inherit_kwargs=True, repr=True
    )
    """The object of the emoji being removed."""


@base_events.requires_intents(intents.Intent.GUILD_MESSAGE_REACTIONS, intents.Intent.DIRECT_MESSAGE_REACTIONS)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class MessageReactionRemoveAllEvent(base_events.HikariEvent, marshaller.Deserializable):
    """Used to represent Message Reaction Remove All gateway events.

    Sent when all the reactions are removed from a message, regardless of emoji.
    """

    channel_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake, repr=True)
    """The ID of the channel where the targeted message is."""

    message_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake, repr=True)
    """The ID of the message all reactions are being removed from."""

    guild_id: typing.Optional[base_entities.Snowflake] = marshaller.attrib(
        deserializer=base_entities.Snowflake, if_undefined=None, default=None, repr=True,
    )
    """The ID of the guild where the targeted message is, if applicable."""


@base_events.requires_intents(intents.Intent.GUILD_MESSAGE_REACTIONS, intents.Intent.DIRECT_MESSAGE_REACTIONS)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class MessageReactionRemoveEmojiEvent(base_events.HikariEvent, marshaller.Deserializable):
    """Represents Message Reaction Remove Emoji events.

    Sent when all the reactions for a single emoji are removed from a message.
    """

    channel_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake, repr=True)
    """The ID of the channel where the targeted message is."""

    guild_id: typing.Optional[base_entities.Snowflake] = marshaller.attrib(
        deserializer=base_entities.Snowflake, if_undefined=None, default=None, repr=True
    )
    """The ID of the guild where the targeted message is, if applicable."""

    message_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake, repr=True)
    """The ID of the message the reactions are being removed from."""

    emoji: typing.Union[emojis.UnicodeEmoji, emojis.UnknownEmoji] = marshaller.attrib(
        deserializer=emojis.deserialize_reaction_emoji, inherit_kwargs=True, repr=True
    )
    """The object of the emoji that's being removed."""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Definition of the interface a compliant weaving implementation should provide.

State object handle decoding events and managing application state.
"""

from __future__ import annotations

__all__ = ["RawEventConsumer"]

import abc
import typing

if typing.TYPE_CHECKING:
    from hikari.clients import shards


class RawEventConsumer(abc.ABC):
    """Consumer of raw events from Discord.

    RawEventConsumer describes an object that takes any event payloads that
    Discord dispatches over a websocket and decides how to process it further.
    This is used as the core base for any form of event manager type.

    This base may also be used by users to dispatch the event to a completely
    different medium, such as a message queue for distributed applications.
    """

    @abc.abstractmethod
    def process_raw_event(
        self, shard_client_obj: shards.ShardClient, name: str, payload: typing.Mapping[str, str],
    ) -> None:
        """Consume a raw event that was received from a shard connection.

        Parameters
        ----------
        shard_client_obj : hikari.clients.shards.ShardClient
            The client for the shard that received the event.
        name : str
            The raw event name.
        payload : typing.Any
            The raw event payload. Will be a JSON-compatible type.
        """



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""State of a shard."""

from __future__ import annotations

__all__ = ["ShardState"]

from hikari.internal import more_enums


@more_enums.must_be_unique
class ShardState(int, more_enums.Enum):
    """Describes the state of a shard."""

    NOT_RUNNING = 0
    """The shard is not running."""

    CONNECTING = more_enums.generated_value()
    """The shard is undergoing the initial connection handshake."""

    WAITING_FOR_READY = more_enums.generated_value()
    """The initialization handshake has completed.

    We are waiting for the shard to receive the `READY` event.
    """

    READY = more_enums.generated_value()
    """The shard is `READY`."""

    RESUMING = more_enums.generated_value()
    """The shard has sent a request to `RESUME` and is waiting for a response."""

    STOPPING = more_enums.generated_value()
    """The shard is currently shutting down permanently."""

    STOPPED = more_enums.generated_value()
    """The shard has shut down and is no longer connected."""

    def __str__(self) -> str:
        return self.name



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Components and entities that are used to describe both custom and Unicode emojis on Discord."""

from __future__ import annotations

__all__ = ["Emoji", "UnicodeEmoji", "UnknownEmoji", "GuildEmoji"]

import typing

import attr

from hikari import bases
from hikari import files
from hikari import users
from hikari.internal import marshaller

if typing.TYPE_CHECKING:
    from hikari.internal import more_typing


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class Emoji(bases.HikariEntity, marshaller.Deserializable):
    """Base class for all emojis."""


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class UnicodeEmoji(Emoji):
    """Represents a unicode emoji."""

    name: str = marshaller.attrib(deserializer=str, repr=True)
    """The codepoints that form the emoji."""

    @property
    def url_name(self) -> str:
        """Get the format of this emoji used in request routes."""
        return self.name

    @property
    def mention(self) -> str:
        """Get the format of this emoji used for sending it in a channel."""
        return self.name


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class UnknownEmoji(Emoji, bases.UniqueEntity, files.BaseStream):
    """Represents a unknown emoji."""

    name: typing.Optional[str] = marshaller.attrib(deserializer=str, if_none=None, repr=True)
    """The name of the emoji."""

    is_animated: typing.Optional[bool] = marshaller.attrib(
        raw_name="animated", deserializer=bool, if_undefined=False, if_none=None, default=False
    )
    """Whether the emoji is animated.

    Will be `None` when received in Message Reaction Remove and Message
    Reaction Remove Emoji events.
    """

    @property
    def url_name(self) -> str:
        """Get the format of this emoji used in request routes."""
        return f"{self.name}:{self.id}"

    @property
    def url(self) -> str:
        # TODO: implement.
        raise NotImplementedError()

    def __aiter__(self) -> typing.AsyncIterator[bytes]:
        # TODO: implement.
        raise NotImplementedError()


def _deserialize_role_ids(payload: more_typing.JSONArray) -> typing.Set[bases.Snowflake]:
    return {bases.Snowflake(role_id) for role_id in payload}


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildEmoji(UnknownEmoji):
    """Represents a guild emoji."""

    role_ids: typing.Set[bases.Snowflake] = marshaller.attrib(
        raw_name="roles", deserializer=_deserialize_role_ids, if_undefined=set, factory=set,
    )
    """The IDs of the roles that are whitelisted to use this emoji.

    If this is empty then any user can use this emoji regardless of their roles.
    """

    user: typing.Optional[users.User] = marshaller.attrib(
        deserializer=users.User.deserialize, if_none=None, if_undefined=None, default=None, inherit_kwargs=True
    )
    """The user that created the emoji.

    !!! note
        This will be `None` if you are missing the `MANAGE_EMOJIS`
        permission in the server the emoji is from.
    """

    is_colons_required: bool = marshaller.attrib(raw_name="require_colons", deserializer=bool)
    """Whether this emoji must be wrapped in colons."""

    is_managed: bool = marshaller.attrib(raw_name="managed", deserializer=bool)
    """Whether the emoji is managed by an integration."""

    is_available: bool = marshaller.attrib(raw_name="available", deserializer=bool)
    """Whether this emoji can currently be used.

    May be `False` due to a loss of Sever Boosts on the emoji's guild.
    """

    @property
    def mention(self) -> str:
        """Get the format of this emoji used for sending it in a channel."""
        return f"<{'a' if self.is_animated else ''}:{self.url_name}>"


def deserialize_reaction_emoji(payload: typing.Dict, **kwargs: typing.Any) -> typing.Union[UnicodeEmoji, UnknownEmoji]:
    """Deserialize a reaction emoji into an emoji."""
    if payload.get("id"):
        return UnknownEmoji.deserialize(payload, **kwargs)

    return UnicodeEmoji.deserialize(payload, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Logic for handling all requests to reaction endpoints."""

from __future__ import annotations

__all__ = ["RESTReactionComponent"]

import abc
import datetime
import functools
import typing

from hikari import bases
from hikari import users
from hikari.clients.rest import base
from hikari.internal import helpers

if typing.TYPE_CHECKING:
    from hikari import channels as _channels
    from hikari import emojis
    from hikari import messages as _messages


class RESTReactionComponent(base.BaseRESTComponent, abc.ABC):  # pylint: disable=abstract-method
    """The REST client component for handling requests to reaction endpoints."""

    async def add_reaction(
        self,
        channel: bases.Hashable[_channels.PartialChannel],
        message: bases.Hashable[_messages.Message],
        emoji: typing.Union[emojis.Emoji, str],
    ) -> None:
        """Add a reaction to the given message in the given channel.

        Parameters
        ----------
        channel : typing.Union[hikari.channels.PartialChannel, hikari.bases.Snowflake, int]
            The object or ID of the channel to add this reaction in.
        message : typing.Union[hikari.messages.Message, hikari.bases.Snowflake, int]
            The object or ID of the message to add the reaction in.
        emoji : typing.Union[hikari.emojis.Emoji, str]
            The emoji to add. This can either be an emoji object or a string
            representation of an emoji. The string representation will be either
            `"name:id"` for custom emojis, or it's unicode character(s) for
            standard emojis.

        Raises
        ------
        hikari.errors.Forbidden
            If this is the first reaction using this specific emoji on this
            message and you lack the `ADD_REACTIONS` permission. If you lack
            `READ_MESSAGE_HISTORY`, this may also raise this error.
        hikari.errors.NotFound
            If the channel or message is not found, or if the emoji is not found.
        hikari.errors.BadRequest
            If the emoji is invalid, unknown, or formatted incorrectly.
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        """
        await self._session.create_reaction(
            channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel)),
            message_id=str(message.id if isinstance(message, bases.UniqueEntity) else int(message)),
            emoji=str(getattr(emoji, "url_name", emoji)),
        )

    async def remove_reaction(
        self,
        channel: bases.Hashable[_channels.PartialChannel],
        message: bases.Hashable[_messages.Message],
        emoji: typing.Union[emojis.Emoji, str],
        *,
        user: typing.Optional[typing.Hashable[users.User]] = None,
    ) -> None:
        """Remove a given reaction from a given message in a given channel.

        If the user is `None`, then the bot's own reaction is removed, otherwise
        the given user's reaction is removed instead.

        Parameters
        ----------
        channel : typing.Union[hikari.channels.PartialChannel, hikari.bases.Snowflake, int]
            The object or ID of the channel to add this reaction in.
        message : typing.Union[hikari.messages.Message, hikari.bases.Snowflake, int]
            The object or ID of the message to add the reaction in.
        emoji : typing.Union[hikari.emojis.Emoji, str]
            The emoji to add. This can either be an emoji object or a
            string representation of an emoji. The string representation will be
            either `"name:id"` for custom emojis else it's unicode
            character(s) (can be UTF-32).
        user : typing.Union[hikari.users.User, hikari.bases.Snowflake, int], optional
            The user to remove the reaction of. If this is `None`, then the
            bot's own reaction is removed instead.

        Raises
        ------
        hikari.errors.Forbidden
            If this is the first reaction using this specific emoji on this
            message and you lack the `ADD_REACTIONS` permission. If you lack
            `READ_MESSAGE_HISTORY`, this may also raise this error.
            If the `user` is not `None` and the bot lacks permissions to
            modify other user's reactions, this will also be raised.
        hikari.errors.NotFound
            If the channel or message is not found, or if the emoji is not
            found.
        hikari.errors.BadRequest
            If the emoji is not valid, unknown, or formatted incorrectly.
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        """
        if user is None:
            await self._session.delete_own_reaction(
                channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel)),
                message_id=str(message.id if isinstance(message, bases.UniqueEntity) else int(message)),
                emoji=str(getattr(emoji, "url_name", emoji)),
            )
        else:
            await self._session.delete_user_reaction(
                channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel)),
                message_id=str(message.id if isinstance(message, bases.UniqueEntity) else int(message)),
                emoji=str(getattr(emoji, "url_name", emoji)),
                user_id=str(user.id if isinstance(user, bases.UniqueEntity) else int(user)),
            )

    async def remove_all_reactions(
        self, channel: bases.Hashable[_channels.PartialChannel], message: bases.Hashable[_messages.Message],
    ) -> None:
        """Delete all reactions from a given message in a given channel.

        Parameters
        ----------
        channel : typing.Union[hikari.channels.PartialChannel, hikari.bases.Snowflake, int]
            The object or ID of the channel to get the message from.
        message : typing.Union[hikari.messages.Message, hikari.bases.Snowflake, int]
            The object or ID of the message to remove all reactions from.

        Raises
        ------
        hikari.errors.BadRequest
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        hikari.errors.NotFound
            If the channel or message is not found.
        hikari.errors.Forbidden
            If you lack the `MANAGE_MESSAGES` permission.
        """
        await self._session.delete_all_reactions(
            channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel)),
            message_id=str(message.id if isinstance(message, bases.UniqueEntity) else int(message)),
        )

    async def remove_all_reactions_for_emoji(
        self,
        channel: bases.Hashable[_channels.PartialChannel],
        message: bases.Hashable[_messages.Message],
        emoji: typing.Union[emojis.Emoji, str],
    ) -> None:
        """Remove all reactions for a single given emoji on a given message.

        Parameters
        ----------
        channel : typing.Union[hikari.channels.PartialChannel, hikari.bases.Snowflake, int]
            The object or ID of the channel to get the message from.
        message : typing.Union[hikari.messages.Message, hikari.bases.Snowflake, int]
            The object or ID of the message to delete the reactions from.
        emoji : typing.Union[hikari.emojis.Emoji, str]
            The object or string representation of the emoji to delete. The
            string representation will be either `"name:id"` for custom emojis
            else it's unicode character(s) (can be UTF-32).

        Raises
        ------
        hikari.errors.BadRequest
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        hikari.errors.NotFound
            If the channel or message or emoji or user is not found.
        hikari.errors.Forbidden
            If you lack the `MANAGE_MESSAGES` permission, or the channel is a
            DM channel.
        """
        await self._session.delete_all_reactions_for_emoji(
            channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel)),
            message_id=str(message.id if isinstance(message, bases.UniqueEntity) else int(message)),
            emoji=str(getattr(emoji, "url_name", emoji)),
        )

    def fetch_reactors_after(
        self,
        channel: bases.Hashable[_channels.PartialChannel],
        message: bases.Hashable[_messages.Message],
        emoji: typing.Union[emojis.Emoji, str],
        *,
        after: typing.Union[datetime.datetime, bases.Hashable[users.User]] = 0,
        limit: typing.Optional[int] = None,
    ) -> typing.AsyncIterator[users.User]:
        """Get an async iterator of the users who reacted to a message.

        This returns the users created after a given user object/ID or from the
        oldest user who reacted.

        Parameters
        ----------
        channel : typing.Union[hikari.channels.PartialChannel, hikari.bases.Snowflake, int]
            The object or ID of the channel to get the message from.
        message : typing.Union[hikari.messages.Message, hikari.bases.Snowflake, int]
            The object or ID of the message to get the reactions from.
        emoji : typing.Union[hikari.emojis.Emoji, str]
            The emoji to get. This can either be it's object or the string
            representation of the emoji. The string representation will be
            either `"name:id"` for custom emojis else it's unicode
            character(s) (can be UTF-32).
        after : typing.Union[datetime.datetime, hikari.users.User, hikari.bases.Snowflake, int]
            If specified, a object or ID user. If specified, only users with a
            snowflake that is lexicographically greater than the value will be
            returned.
        limit : str
            If specified, the limit of the number of users this iterator should
            return.

        Examples
        --------
            async for user in client.fetch_reactors_after(channel, message, emoji, after=9876543, limit=1231):
                if user.is_bot:
                    await client.kick_member(channel.guild_id, user)

        Returns
        -------
        typing.AsyncIterator[hikari.users.User]
            An async iterator of user objects.

        Raises
        ------
        hikari.errors.BadRequest
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        hikari.errors.Forbidden
            If you lack access to the message.
        hikari.errors.NotFound
            If the channel or message is not found.
        """
        if isinstance(after, datetime.datetime):
            after = str(bases.Snowflake.from_datetime(after))
        else:
            after = str(after.id if isinstance(after, bases.UniqueEntity) else int(after))
        request = functools.partial(
            self._session.get_reactions,
            channel_id=str(channel.id if isinstance(channel, bases.UniqueEntity) else int(channel)),
            message_id=str(message.id if isinstance(message, bases.UniqueEntity) else int(message)),
            emoji=getattr(emoji, "url_name", emoji),
        )
        deserializer = functools.partial(users.User.deserialize, components=self._components)
        return helpers.pagination_handler(
            deserializer=deserializer,
            direction="after",
            request=request,
            reversing=False,
            start=after,
            maximum_limit=100,
            limit=limit,
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Components and entities that are used to describe Users on Discord."""

from __future__ import annotations

__all__ = ["User", "MyUser", "UserFlag", "PremiumType"]

import typing

import attr

from hikari import bases
from hikari.internal import marshaller
from hikari.internal import more_enums
from hikari.internal import urls


@more_enums.must_be_unique
class UserFlag(more_enums.IntFlag):
    """The known user flags that represent account badges."""

    NONE = 0
    """None"""

    DISCORD_EMPLOYEE = 1 << 0
    """Discord Employee"""

    DISCORD_PARTNER = 1 << 1
    """Discord Partner"""

    HYPESQUAD_EVENTS = 1 << 2
    """HypeSquad Events"""

    BUG_HUNTER_LEVEL_1 = 1 << 3
    """Bug Hunter Level 1"""

    HOUSE_BRAVERY = 1 << 6
    """House of Bravery"""

    HOUSE_BRILLIANCE = 1 << 7
    """House of Brilliance"""

    HOUSE_BALANCE = 1 << 8
    """House of Balance"""

    EARLY_SUPPORTER = 1 << 9
    """Early Supporter"""

    TEAM_USER = 1 << 10
    """Team user"""

    SYSTEM = 1 << 12
    """System"""

    BUG_HUNTER_LEVEL_2 = 1 << 14
    """Bug Hunter Level 2"""

    VERIFIED_BOT = 1 << 16
    """Verified Bot"""

    VERIFIED_BOT_DEVELOPER = 1 << 17
    """Verified Bot Developer"""


@more_enums.must_be_unique
class PremiumType(int, more_enums.Enum):
    """The types of Nitro."""

    NONE = 0
    """No premium."""

    NITRO_CLASSIC = 1
    """Premium including basic perks like animated emojis and avatars."""

    NITRO = 2
    """Premium including all perks (e.g. 2 server boosts)."""


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class User(bases.UniqueEntity, marshaller.Deserializable):
    """Represents a user."""

    discriminator: str = marshaller.attrib(deserializer=str, repr=True)
    """This user's discriminator."""

    username: str = marshaller.attrib(deserializer=str, repr=True)
    """This user's username."""

    avatar_hash: typing.Optional[str] = marshaller.attrib(raw_name="avatar", deserializer=str, if_none=None)
    """This user's avatar hash, if set."""

    is_bot: bool = marshaller.attrib(raw_name="bot", deserializer=bool, if_undefined=False, default=False)
    """Whether this user is a bot account."""

    is_system: bool = marshaller.attrib(raw_name="system", deserializer=bool, if_undefined=False, default=False)
    """Whether this user is a system account."""

    flags: typing.Optional[UserFlag] = marshaller.attrib(
        raw_name="public_flags", deserializer=UserFlag, if_undefined=None, default=None
    )
    """The public flags for this user.

    !!! info
        This will be `None` if it's a webhook user.
    """

    @property
    def avatar_url(self) -> str:
        """URL for this user's custom avatar if set, else default."""
        return self.format_avatar_url()

    def format_avatar_url(self, fmt: typing.Optional[str] = None, size: int = 4096) -> str:
        """Generate the avatar URL for this user's custom avatar if set, else their default avatar.

        Parameters
        ----------
        fmt : str
            The format to use for this URL, defaults to `png` or `gif`.
            Supports `png`, `jpeg`, `jpg`, `webp` and `gif` (when
            animated). Will be ignored for default avatars which can only be
            `png`.
        size : int
            The size to set for the URL, defaults to `4096`.
            Can be any power of two between 16 and 4096.
            Will be ignored for default avatars.

        Returns
        -------
        str
            The string URL.

        Raises
        ------
        ValueError
            If `size` is not a power of two or not between 16 and 4096.
        """
        if not self.avatar_hash:
            return urls.generate_cdn_url("embed/avatars", str(self.default_avatar), fmt="png", size=None)
        if fmt is None and self.avatar_hash.startswith("a_"):
            fmt = "gif"
        elif fmt is None:
            fmt = "png"
        return urls.generate_cdn_url("avatars", str(self.id), self.avatar_hash, fmt=fmt, size=size)

    @property
    def default_avatar(self) -> int:
        """Integer representation of this user's default avatar."""
        return int(self.discriminator) % 5


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class MyUser(User):
    """Represents a user with extended oauth2 information."""

    is_mfa_enabled: bool = marshaller.attrib(raw_name="mfa_enabled", deserializer=bool)
    """Whether the user's account has 2fa enabled."""

    locale: typing.Optional[str] = marshaller.attrib(deserializer=str, if_none=None, if_undefined=None, default=None)
    """The user's set language. This is not provided by the `READY` event."""

    is_verified: typing.Optional[bool] = marshaller.attrib(
        raw_name="verified", deserializer=bool, if_undefined=None, default=None
    )
    """Whether the email for this user's account has been verified.

    Will be `None` if retrieved through the oauth2 flow without the `email`
    scope.
    """

    email: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None, if_none=None, default=None)
    """The user's set email.

    Will be `None` if retrieved through the oauth2 flow without the `email`
    scope and for bot users.
    """

    flags: UserFlag = marshaller.attrib(deserializer=UserFlag)
    """This user account's flags."""

    premium_type: typing.Optional[PremiumType] = marshaller.attrib(
        deserializer=PremiumType, if_undefined=None, default=None
    )
    """The type of Nitro Subscription this user account had.

    This will always be `None` for bots.
    """



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Bases for components and entities that are used to describe Discord gateway events."""
from __future__ import annotations

__all__ = ["HikariEvent", "get_required_intents_for", "requires_intents", "no_catch", "is_no_catch_event"]

import abc
import typing

import attr

from hikari import bases
from hikari.internal import marshaller
from hikari.internal import more_collections

if typing.TYPE_CHECKING:
    from hikari import intents


# Base event, is not deserialized
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class HikariEvent(bases.HikariEntity, abc.ABC):
    """The base class that all events inherit from."""


_HikariEventT = typing.TypeVar("_HikariEventT", contravariant=True)
_REQUIRED_INTENTS_ATTR: typing.Final[str] = "___required_intents___"
_NO_THROW_ATTR: typing.Final[str] = "___no_throw___"


def get_required_intents_for(event_type: typing.Type[HikariEvent]) -> typing.Collection[intents.Intent]:
    """Retrieve the intents that are required to listen to an event type.

    Parameters
    ----------
    event_type : typing.Type[HikariEvent]
        The event type to get required intents for.

    Returns
    -------
    typing.Collection[hikari.intents.Intent]
        Collection of acceptable subset combinations of intent needed to
        be able to receive the given event type.
    """
    return getattr(event_type, _REQUIRED_INTENTS_ATTR, more_collections.EMPTY_COLLECTION)


def requires_intents(
    first: intents.Intent, *rest: intents.Intent
) -> typing.Callable[[typing.Type[_HikariEventT]], typing.Type[_HikariEventT]]:
    """Decorate an event type to define what intents it requires.

    Parameters
    ----------
    first : hikari.intents.Intent
        First combination of intents that are acceptable in order to receive
        the decorated event type.
    *rest : hikari.intents.Intent
        Zero or more additional combinations of intents to require for this
        event to be subscribed to.

    """

    def decorator(cls: typing.Type[_HikariEventT]) -> typing.Type[_HikariEventT]:
        setattr(cls, _REQUIRED_INTENTS_ATTR, [first, *rest])
        return cls

    return decorator


def no_catch():
    """Decorate an event type to indicate errors should not be handled.

    This is useful for exception event types that you do not want to
    have invoked recursively.
    """

    def decorator(cls: typing.Type[_HikariEventT]) -> typing.Type[_HikariEventT]:
        setattr(cls, _NO_THROW_ATTR, True)
        return cls

    return decorator


def is_no_catch_event(obj: typing.Union[_HikariEventT, typing.Type[_HikariEventT]]) -> bool:
    """Return True if this event is marked as `no_catch`."""
    return getattr(obj, _NO_THROW_ATTR, False)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Core errors that may be raised by this API implementation."""

from __future__ import annotations

__all__ = [
    "HikariError",
    "HikariWarning",
    "NotFound",
    "Unauthorized",
    "Forbidden",
    "BadRequest",
    "HTTPError",
    "HTTPErrorResponse",
    "ClientHTTPErrorResponse",
    "ServerHTTPErrorResponse",
    "GatewayZombiedError",
    "GatewayNeedsShardingError",
    "GatewayMustReconnectError",
    "GatewayInvalidSessionError",
    "GatewayInvalidTokenError",
    "GatewayServerClosedConnectionError",
    "GatewayClientClosedError",
    "GatewayClientDisconnectedError",
    "GatewayError",
]

import http
import typing

import aiohttp.typedefs

from hikari.net import codes


class HikariError(RuntimeError):
    """Base for an error raised by this API.

    Any exceptions should derive from this.

    !!! note
        You should never initialize this exception directly.
    """

    __slots__ = ()

    def __repr__(self):
        return str(self)


class HikariWarning(RuntimeWarning):
    """Base for a warning raised by this API.

    Any warnings should derive from this.

    !!! note
        You should never initialize this warning directly.
    """

    __slots__ = ()


class GatewayError(HikariError):
    """A base exception type for anything that can be thrown by the Gateway.

    Parameters
    ----------
    reason : str
        A string explaining the issue.
    """

    __slots__ = ("reason",)

    reason: str
    """A string to explain the issue."""

    def __init__(self, reason: str) -> None:
        super().__init__()
        self.reason = reason

    def __str__(self) -> str:
        return self.reason


class GatewayClientClosedError(GatewayError):
    """An exception raised when you programmatically shut down the bot client-side.

    Parameters
    ----------
    reason : str
        A string explaining the issue.
    """

    __slots__ = ()

    def __init__(self, reason: str = "The gateway client has been closed") -> None:
        super().__init__(reason)


class GatewayClientDisconnectedError(GatewayError):
    """An exception raised when the bot client-side disconnects unexpectedly.

    Parameters
    ----------
    reason : str
        A string explaining the issue.
    """

    __slots__ = ()

    def __init__(self, reason: str = "The gateway client has disconnected unexpectedly") -> None:
        super().__init__(reason)


class GatewayServerClosedConnectionError(GatewayError):
    """An exception raised when the server closes the connection.

    Parameters
    ----------
    close_code : typing.Union[hikari.net.codes.GatewayCloseCode, int], optional
        The close code provided by the server, if there was one.
    reason : str, optional
        A string explaining the issue.
    """

    __slots__ = ("close_code",)

    close_code: typing.Union[codes.GatewayCloseCode, int, None]

    def __init__(
        self,
        close_code: typing.Optional[typing.Union[codes.GatewayCloseCode, int]] = None,
        reason: typing.Optional[str] = None,
    ) -> None:
        try:
            name = close_code.name
        except AttributeError:
            name = str(close_code) if close_code is not None else "no reason"

        if reason is None:
            reason = f"Gateway connection closed by server ({name})"

        self.close_code = close_code
        super().__init__(reason)


class GatewayInvalidTokenError(GatewayServerClosedConnectionError):
    """An exception that is raised if you failed to authenticate with a valid token to the Gateway."""

    __slots__ = ()

    def __init__(self) -> None:
        super().__init__(
            codes.GatewayCloseCode.AUTHENTICATION_FAILED,
            "The account token specified is invalid for the gateway connection",
        )


class GatewayInvalidSessionError(GatewayServerClosedConnectionError):
    """An exception raised if a Gateway session becomes invalid.

    Parameters
    ----------
    can_resume : bool
        `True` if the connection will be able to RESUME next time it starts
        rather than re-IDENTIFYing, or `False` if you need to IDENTIFY
        again instead.
    """

    __slots__ = ("can_resume",)

    can_resume: bool
    """`True` if the next reconnection can be RESUMED,
    `False` if it has to be coordinated by re-IDENFITYing.
    """

    def __init__(self, can_resume: bool) -> None:
        self.can_resume = can_resume
        instruction = "restart the shard and RESUME" if can_resume else "restart the shard with a fresh session"
        super().__init__(reason=f"The session has been invalidated; {instruction}")


class GatewayMustReconnectError(GatewayServerClosedConnectionError):
    """An exception raised when the Gateway has to re-connect with a new session.

    This will cause a re-IDENTIFY.
    """

    __slots__ = ()

    def __init__(self) -> None:
        super().__init__(reason="The gateway server has requested that the client reconnects with a new session")


class GatewayNeedsShardingError(GatewayServerClosedConnectionError):
    """An exception raised if you have too many guilds on one of the current Gateway shards.

    This is a sign you need to increase the number of shards that your bot is
    running with in order to connect to Discord.
    """

    __slots__ = ()

    def __init__(self) -> None:
        super().__init__(
            codes.GatewayCloseCode.SHARDING_REQUIRED, "You are in too many guilds. Shard the bot to connect",
        )


class GatewayZombiedError(GatewayClientClosedError):
    """An exception raised if a shard becomes zombied.

    This means that Discord is no longer responding to us, and we have
    disconnected due to a timeout.
    """

    __slots__ = ()

    def __init__(self) -> None:
        super().__init__("No heartbeat was received, the connection has been closed")


class HTTPError(HikariError):
    """Base exception raised if an HTTP error occurs while making a request.

    Parameters
    ----------
    message : str
        The error message.
    url : str
        The URL that produced this error.
    """

    __slots__ = ("message", "url")

    message: str
    """The error message."""

    url: str
    """The URL that produced this error message."""

    def __init__(self, url: str, message: str) -> None:
        super().__init__()
        self.message = message
        self.url = url


class HTTPErrorResponse(HTTPError):
    """Base exception for an erroneous HTTP response.

    Parameters
    ----------
    url : str
        The URL that produced the error message.
    status : http.HTTPStatus
        The HTTP status code of the response that caused this error.
    headers : aiohttp.typedefs.LooseHeaders
        Any headers that were given in the response.
    raw_body : typing.Any
        The body that was received.
    """

    __slots__ = ("status", "headers", "raw_body")

    status: http.HTTPStatus
    """The HTTP status code for the response."""

    headers: aiohttp.typedefs.LooseHeaders
    """The headers received in the error response."""

    raw_body: typing.Any
    """The response body."""

    def __init__(
        self, url: str, status: http.HTTPStatus, headers: aiohttp.typedefs.LooseHeaders, raw_body: typing.Any,
    ) -> None:
        super().__init__(url, f"{status}: {raw_body}")
        self.status = status
        self.headers = headers
        self.raw_body = raw_body

    def __str__(self) -> str:
        return f"{self.status.value} {self.status.name}: {self.raw_body}"


class ClientHTTPErrorResponse(HTTPErrorResponse):
    """Base exception for an erroneous HTTP response that is a client error.

    All exceptions derived from this base should be treated as 4xx client
    errors when encountered.
    """

    __slots__ = ()


class BadRequest(ClientHTTPErrorResponse):
    """Raised when you send an invalid request somehow.

    Parameters
    ----------
    url : str
        The URL that produced the error message.
    headers : aiohttp.typedefs.LooseHeaders
        Any headers that were given in the response.
    raw_body : typing.Any
        The body that was received.
    """

    __slots__ = ()

    def __init__(self, url: str, headers: aiohttp.typedefs.LooseHeaders, raw_body: typing.AnyStr) -> None:
        status = http.HTTPStatus.BAD_REQUEST
        super().__init__(url, status, headers, raw_body)


class Unauthorized(ClientHTTPErrorResponse):
    """Raised when you are not authorized to access a specific resource.

    This generally means you did not provide a token, or the token is invalid.

    Parameters
    ----------
    url : str
        The URL that produced the error message.
    headers : aiohttp.typedefs.LooseHeaders
        Any headers that were given in the response.
    raw_body : typing.Any
        The body that was received.
    """

    __slots__ = ()

    def __init__(self, url: str, headers: aiohttp.typedefs.LooseHeaders, raw_body: typing.AnyStr) -> None:
        status = http.HTTPStatus.UNAUTHORIZED
        super().__init__(url, status, headers, raw_body)


class Forbidden(ClientHTTPErrorResponse):
    """Raised when you are not allowed to access a specific resource.

    This means you lack the permissions to do something, either because of
    permissions set in a guild, or because your application is not whitelisted
    to use a specific endpoint.

    Parameters
    ----------
    url : str
        The URL that produced the error message.
    headers : aiohttp.typedefs.LooseHeaders
        Any headers that were given in the response.
    raw_body : typing.Any
        The body that was received.
    """

    __slots__ = ()

    def __init__(self, url: str, headers: aiohttp.typedefs.LooseHeaders, raw_body: typing.AnyStr) -> None:
        status = http.HTTPStatus.FORBIDDEN
        super().__init__(url, status, headers, raw_body)


class NotFound(ClientHTTPErrorResponse):
    """Raised when something is not found.

    Parameters
    ----------
    url : str
        The URL that produced the error message.
    headers : aiohttp.typedefs.LooseHeaders
        Any headers that were given in the response.
    raw_body : typing.Any
        The body that was received.
    """

    __slots__ = ()

    def __init__(self, url: str, headers: aiohttp.typedefs.LooseHeaders, raw_body: typing.AnyStr) -> None:
        status = http.HTTPStatus.NOT_FOUND
        super().__init__(url, status, headers, raw_body)


class ServerHTTPErrorResponse(HTTPErrorResponse):
    """Base exception for an erroneous HTTP response that is a server error.

    All exceptions derived from this base should be treated as 5xx server
    errors when encountered. If you get one of these, it isn't your fault!
    """

    __slots__ = ()


class IntentWarning(HikariWarning):
    """Warning raised when subscribing to an event that cannot be fired.

    This is caused by your application missing certain intents.
    """

    __slots__ = ()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""The logic for handling all requests to user endpoints."""

from __future__ import annotations

__all__ = ["RESTUserComponent"]

import abc

from hikari import bases
from hikari import users
from hikari.clients.rest import base


class RESTUserComponent(base.BaseRESTComponent, abc.ABC):  # pylint: disable=abstract-method
    """The REST client component for handling requests to user endpoints."""

    async def fetch_user(self, user: bases.Hashable[users.User]) -> users.User:
        """Get a given user.

        Parameters
        ----------
        user : typing.Union[hikari.users.User, hikari.bases.Snowflake, int]
            The object or ID of the user to get.

        Returns
        -------
        hikari.users.User
            The requested user object.

        Raises
        ------
        hikari.errors.BadRequest
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        hikari.errors.NotFound
            If the user is not found.
        """
        payload = await self._session.get_user(
            user_id=str(user.id if isinstance(user, bases.UniqueEntity) else int(user))
        )
        return users.User.deserialize(payload, components=self._components)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Asyncio extensions and utilities."""

from __future__ import annotations

__all__ = ["completed_future", "wait", "is_async_iterator", "is_async_iterable"]

import asyncio
import inspect
import typing

if typing.TYPE_CHECKING:
    from hikari.internal import more_typing


@typing.overload
def completed_future() -> more_typing.Future[None]:
    """Return a completed future with no result."""


@typing.overload
def completed_future(result: more_typing.T_contra, /) -> more_typing.Future[more_typing.T_contra]:
    """Return a completed future with the given value as the result."""


def completed_future(result=None, /):
    """Create a future on the current running loop that is completed, then return it.

    Parameters
    ----------
    result : typing.Any
        The value to set for the result of the future.

    Returns
    -------
    asyncio.Future
        The completed future.
    """
    future = asyncio.get_event_loop().create_future()
    future.set_result(result)
    return future


def wait(
    aws: typing.Union[more_typing.Coroutine[more_typing.T_co], typing.Awaitable],
    *,
    timeout=None,
    return_when=asyncio.ALL_COMPLETED,
) -> more_typing.Coroutine[
    typing.Tuple[typing.Set[more_typing.Future[more_typing.T_co]], typing.Set[more_typing.Future[more_typing.T_co]]]
]:
    """Run awaitable objects in the aws set concurrently.

    This blocks until the condition specified by `return_value`.

    Returns
    -------
    typing.Tuple with two typing.Set of futures
        The coroutine returned by `asyncio.wait` of two sets of
        Tasks/Futures (done, pending).
    """
    # noinspection PyTypeChecker
    return asyncio.wait([asyncio.ensure_future(f) for f in aws], timeout=timeout, return_when=return_when)


# On Python3.8.2, there appears to be a bug with the typing module:

# >>> class Aiterable:
# ...     async def __aiter__(self):
# ...         yield ...
# >>> isinstance(Aiterable(), typing.AsyncIterable)
# True

# >>> class Aiterator:
# ...     async def __anext__(self):
# ...         return ...
# >>> isinstance(Aiterator(), typing.AsyncIterator)
# False

# ... so I guess I will have to determine this some other way.


def is_async_iterator(obj: typing.Any) -> bool:
    """Determine if the object is an async iterator or not."""
    return hasattr(obj, "__anext__") and asyncio.iscoroutinefunction(obj.__anext__)


def is_async_iterable(obj: typing.Any) -> bool:
    """Determine if the object is an async iterable or not."""
    if not hasattr(obj, "__aiter__"):
        return False
    # These could be async generators, or they could be something different.
    return inspect.isfunction(obj.__aiter__) or inspect.ismethod(obj.__aiter__)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Event dispatcher implementations that are intent-aware."""
from __future__ import annotations

__all__ = ["IntentAwareEventDispatcherImpl"]

import asyncio
import logging
import typing

from hikari import errors
from hikari import intents
from hikari.events import base
from hikari.events import other
from hikari.internal import assertions
from hikari.internal import helpers
from hikari.internal import more_asyncio
from hikari.internal import more_collections
from hikari.state import dispatchers

if typing.TYPE_CHECKING:
    from hikari.internal import more_typing


class IntentAwareEventDispatcherImpl(dispatchers.EventDispatcher):
    """Handles storing and dispatching to event listeners and one-time event waiters.

    Event listeners once registered will be stored until they are manually
    removed. Each time an event is dispatched with a matching name, they will
    be invoked on the event loop.

    One-time event waiters are futures that will be completed when a matching
    event is fired. Once they are matched, they are removed from the listener
    list. Each listener has a corresponding predicate that is invoked prior
    to completing the waiter, with any event parameters being passed to the
    predicate. If the predicate returns False, the waiter is not completed. This
    allows filtering of certain events and conditions in a procedural way.

    Events that require a specific intent will trigger warnings on subscription
    if the provided enabled intents are not a superset of this.

    Parameters
    ----------
    enabled_intents : hikari.intents.Intent, optional
        The intents that are enabled for the application. If `None`, then no
        intent checks are performed when subscribing a new event.
    """

    logger: logging.Logger
    """The logger used to write log messages."""

    def __init__(self, enabled_intents: typing.Optional[intents.Intent]) -> None:
        self._enabled_intents = enabled_intents
        self._listeners = {}
        self._waiters = {}
        self.logger = logging.getLogger(type(self).__qualname__)

    def close(self) -> None:
        """Cancel anything that is waiting for an event to be dispatched."""
        self._listeners.clear()
        for waiter in self._waiters.values():
            for future in waiter.keys():
                future.cancel()
        self._waiters.clear()

    def add_listener(
        self, event_type: typing.Type[base.HikariEvent], callback: dispatchers.EventCallbackT, **kwargs
    ) -> dispatchers.EventCallbackT:
        """Register a new event callback to a given event name.

        Parameters
        ----------
        event_type : typing.Type[hikari.events.HikariEvent]
            The event to register to.
        callback : `async def callback(event: HikariEvent) -> ...`
            The event callback to invoke when this event is fired.

        Returns
        -------
        async (event) -> None
            The async callback that was registered.

        Raises
        ------
        TypeError
            If `coroutine_function` is not a coroutine, or if the passed event
            type does not subclass `hikari.events.HikariEvent`.

        Note
        ----
        If you subscribe to an event that requires intents that you do not have
        set, you will receive a warning.
        """
        assertions.assert_that(
            asyncio.iscoroutinefunction(callback), "You must subscribe a coroutine function only", TypeError
        )

        assertions.assert_that(
            issubclass(event_type, base.HikariEvent), "Events must subclass hikari.events.HikariEvent", TypeError
        )

        required_intents = base.get_required_intents_for(event_type)
        enabled_intents = self._enabled_intents if self._enabled_intents is not None else 0

        any_intent_match = any(enabled_intents & i == i for i in required_intents)

        if self._enabled_intents is not None and required_intents and not any_intent_match:
            intents_lists = []
            for required in required_intents:
                set_of_intents = []
                for intent in intents.Intent:
                    if required & intent:
                        set_of_intents.append(f"{intent.name} <PRIVILEGED>" if intent.is_privileged else intent.name)
                intents_lists.append(" + ".join(set_of_intents))

            message = (
                f"Event {event_type.__module__}.{event_type.__qualname__} will never be triggered\n"
                f"unless you enable one of the following intents:\n"
                + "\n".join(f"    - {intent_list}" for intent_list in intents_lists)
            )

            helpers.warning(message, category=errors.IntentWarning, stack_level=kwargs.pop("_stack_level", 1))

        if event_type not in self._listeners:
            self._listeners[event_type] = []
        self._listeners[event_type].append(callback)

        return callback

    def remove_listener(
        self, event_type: typing.Type[dispatchers.EventT], callback: dispatchers.EventCallbackT
    ) -> None:
        """Remove the given coroutine function from the handlers for the given event.

        The name is mandatory to enable supporting registering the same event callback for multiple event types.

        Parameters
        ----------
        event_type : typing.Type[hikari.events.HikariEvent]
            The type of event to remove the callback from.
        callback : `async def callback(event: HikariEvent) -> ...`
            The event callback to remove.
        """
        if event_type in self._listeners and callback in self._listeners[event_type]:
            if len(self._listeners[event_type]) - 1 == 0:
                del self._listeners[event_type]
            else:
                self._listeners[event_type].remove(callback)

    # Do not add an annotation here, it will mess with type hints in PyCharm which can lead to
    # confusing telepathy comments to the user.
    # Additionally, this MUST NOT BE A COROUTINE ITSELF. THIS IS NOT TYPESAFE!
    def dispatch_event(self, event: base.HikariEvent):
        """Dispatch a given event to all listeners and waiters that are applicable.

        Parameters
        ----------
        event : hikari.events.HikariEvent
            The event to dispatch.

        Returns
        -------
        asyncio.Future
            This may be a gathering future of the callbacks to invoke, or it may
            be a completed future object. Regardless, this result will be
            scheduled on the event loop automatically, and does not need to be
            awaited. Awaiting this future will await completion of all invoked
            event handlers.
        """
        this_event_type = type(event)
        self.logger.debug("dispatching %s", this_event_type.__name__)

        callback_futures = []

        for base_event_type in this_event_type.mro():
            for callback in self._listeners.get(base_event_type, more_collections.EMPTY_COLLECTION):
                callback_futures.append(asyncio.create_task(self._failsafe_invoke(event, callback)))

            if base_event_type not in self._waiters:
                continue

            # Quicker most of the time to iterate twice, than to copy the entire collection
            # to iterate once after that.
            futures_to_remove = []

            subtype_waiters = self._waiters.get(base_event_type, more_collections.EMPTY_DICT)

            for future, predicate in subtype_waiters.items():
                # We execute async predicates differently to sync, because we hope most of the time
                # these checks will be synchronous only, as these will perform significantly faster.
                # I preferred execution speed over terseness here.
                if asyncio.iscoroutinefunction(predicate):
                    # Reawaken it later once the predicate is complete. We can await this with the
                    # other dispatchers.
                    check_task = asyncio.create_task(self._async_check(future, predicate, event, base_event_type))
                    callback_futures.append(check_task)
                else:
                    try:
                        if predicate(event):
                            # We have to append this to a list, we can't mutate the dict while we iterate over it...
                            future.set_result(event)
                            futures_to_remove.append(future)
                    except Exception as ex:  # pylint:disable=broad-except
                        future.set_exception(ex)
                        futures_to_remove.append(future)

            # We do this after to prevent changes to the dict while iterating causing exceptions.
            for future in futures_to_remove:
                # Off to the garbage collector you go.
                subtype_waiters.pop(future)

            # If there are no waiters left, purge the entire dict.
            if not subtype_waiters:
                self._waiters.pop(base_event_type)

        result = asyncio.gather(*callback_futures) if callback_futures else more_asyncio.completed_future()

        # Stop intellij shenanigans with broken type hints that ruin my day.
        return typing.cast(typing.Any, result)

    async def _async_check(self, future, predicate, event, event_type):
        # If the predicate returns true, complete the future and pop it from the waiters.
        # By this point we shouldn't be iterating over it anymore, so this is concurrent-modification
        # safe on a single event loop.
        try:
            if await predicate(event):
                future.set_result(event)
                self._waiters[event_type].pop(future)
        except Exception as ex:  # pylint:disable=broad-except
            future.set_exception(ex)
            self._waiters[event_type].pop(future)

    async def _failsafe_invoke(self, event, callback):
        try:
            await callback(event)
        except Exception as ex:  # pylint:disable=broad-except
            self.handle_exception(ex, event, callback)

    def handle_exception(
        self,
        exception: Exception,
        event: base.HikariEvent,
        callback: typing.Callable[..., typing.Union[typing.Awaitable[None]]],
    ) -> None:
        """Handle raised exception.

        This allows users to override this with a custom implementation if desired.

        This implementation will check to see if the event that triggered the
        exception is an `hikari.events.ExceptionEvent`. If this
        exception was caused by the `hikari.events.ExceptionEvent`,
        then nothing is dispatched (thus preventing an exception handler recursively
        re-triggering itself). Otherwise, an `hikari.events.ExceptionEvent`
        is dispatched.

        Parameters
        ----------
        exception: Exception
            The exception that triggered this call.
        event: hikari.events.HikariEvent
            The event that was being dispatched.
        callback
            The callback that threw the exception. This may be an event
            callback, or a `wait_for` predicate that threw an exception.
        """
        # Do not recurse if a dodgy exception handler is added.
        if not base.is_no_catch_event(event):
            self.logger.exception(
                'Exception occurred in handler for event "%s"', type(event).__name__, exc_info=exception
            )
            self.dispatch_event(other.ExceptionEvent(exception=exception, event=event, callback=callback))
        else:
            self.logger.exception(
                'Exception occurred in handler for event "%s", and the exception has been dropped',
                type(event).__name__,
                exc_info=exception,
            )

    def wait_for(
        self,
        event_type: typing.Type[dispatchers.EventT],
        *,
        timeout: typing.Optional[float],
        predicate: dispatchers.PredicateT,
    ) -> more_typing.Future:
        """Wait for a event to occur once and then return the arguments the event was called with.

        Events can be filtered using a given predicate function. If unspecified,
        the first event of the given name will be a match.

        Every event that matches the event name that the bot receives will be
        checked. Thus, if you need to wait for events in a specific guild or
        channel, or from a specific person, you want to give a predicate that
        checks this.

        Parameters
        ----------
        event_type : typing.Type[hikari.events.HikariEvent]
            The name of the event to wait for.
        timeout : float, optional
            The timeout to wait for before cancelling and raising an
            `asyncio.TimeoutError` instead. If this is `None`, this will
            wait forever. Care must be taken if you use `None` as this may
            leak memory if you do this from an event listener that gets
            repeatedly called. If you want to do this, you should consider
            using an event listener instead of this function.
        predicate : `def predicate(event) -> bool`
            A function that takes the arguments for the event and returns True
            if it is a match, or False if it should be ignored. This must be
            a regular function.

        Returns
        -------
        asyncio.Future
            A future that when awaited will provide a the arguments passed to
            the first matching event. If no arguments are passed to the event,
            then `None` is the result. If one argument is passed to the event,
            then that argument is the result, otherwise a tuple of arguments is
            the result instead.

        !!! note
            Awaiting this result will raise an `asyncio.TimeoutError` if the
            timeout is hit and no match is found. If the predicate throws any
            exception, this is raised immediately.
        """
        future = asyncio.get_event_loop().create_future()
        if event_type not in self._waiters:
            # This is used as a weakref dict to allow automatically tidying up
            # any future that falls out of scope entirely.
            self._waiters[event_type] = more_collections.WeakKeyDictionary()
        self._waiters[event_type][future] = predicate
        # noinspection PyTypeChecker
        return asyncio.ensure_future(asyncio.wait_for(future, timeout))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""General helper functions and classes that are not categorised elsewhere."""

from __future__ import annotations

__all__ = ["warning"]

import textwrap
import typing
import warnings

from hikari import bases
from hikari.internal import assertions
from hikari.internal import more_collections

if typing.TYPE_CHECKING:
    from hikari import guilds
    from hikari import users
    from hikari.internal import more_typing


def warning(message: str, category: typing.Type[Warning], stack_level: int = 1) -> None:
    """Generate a warning in a style consistent for this library.

    Parameters
    ----------
    message : str
        The message to display.
    category : typing.Type[Warning]
        The type of warning to raise.
    stack_level : int
        How many stack frames to go back to find the user's invocation.

    """
    warnings.warn("\n\n" + textwrap.indent(message, " " * 2), category, stacklevel=stack_level + 1)


def generate_allowed_mentions(  # pylint:disable=line-too-long
    mentions_everyone: bool,
    user_mentions: typing.Union[typing.Collection[bases.Hashable[users.User]], bool],
    role_mentions: typing.Union[typing.Collection[bases.Hashable[guilds.GuildRole]], bool],
) -> typing.Dict[str, typing.Sequence[str]]:
    """Generate an allowed mentions object based on input mention rules.

    Parameters
    ----------
    mentions_everyone : bool
        Whether `@everyone` and `@here` mentions should be resolved by
        discord and lead to actual pings.
    user_mentions : typing.Union[typing.Collection[typing.Union[hikari.users.User, hikari.bases.Snowflake, int]], bool]
        Either an array of user objects/IDs to allow mentions for,
        `True` to allow all user mentions or `False` to block all
        user mentions from resolving.
    role_mentions : typing.Union[typing.Collection[typing.Union[hikari.guilds.GuildRole, hikari.bases.Snowflake, int]], bool]
        Either an array of guild role objects/IDs to allow mentions for,
        `True` to allow all role mentions or `False` to block all
        role mentions from resolving.

    Returns
    -------
    typing.Dict[str, typing.Sequence[str]]
        The resulting allowed mentions dict object.

    Raises
    ------
    ValueError
        If more than 100 unique objects/entities are passed for
        `role_mentions` or `user_mentions.
    """
    parsed_mentions = []
    allowed_mentions = {}
    if mentions_everyone is True:
        parsed_mentions.append("everyone")
    if user_mentions is True:
        parsed_mentions.append("users")
    # This covers both `False` and an array of IDs/objs by using `user_mentions or EMPTY_SEQUENCE`, where a
    # resultant empty list will mean that all user mentions are blacklisted.
    else:
        allowed_mentions["users"] = list(
            # dict.fromkeys is used to remove duplicate entries that would cause discord to return an error.
            dict.fromkeys(
                str(user.id if isinstance(user, bases.UniqueEntity) else int(user))
                for user in user_mentions or more_collections.EMPTY_SEQUENCE
            )
        )
        assertions.assert_that(len(allowed_mentions["users"]) <= 100, "Only up to 100 users can be provided.")
    if role_mentions is True:
        parsed_mentions.append("roles")
    # This covers both `False` and an array of IDs/objs by using `user_mentions or EMPTY_SEQUENCE`, where a
    # resultant empty list will mean that all role mentions are blacklisted.
    else:
        allowed_mentions["roles"] = list(
            # dict.fromkeys is used to remove duplicate entries that would cause discord to return an error.
            dict.fromkeys(
                str(role.id if isinstance(role, bases.UniqueEntity) else int(role))
                for role in role_mentions or more_collections.EMPTY_SEQUENCE
            )
        )
        assertions.assert_that(len(allowed_mentions["roles"]) <= 100, "Only up to 100 roles can be provided.")
    allowed_mentions["parse"] = parsed_mentions
    # As a note, discord will also treat an empty `allowed_mentions` object as if it wasn't passed at all, so we
    # want to use empty lists for blacklisting elements rather than just not including blacklisted elements.
    return allowed_mentions


# pylint: disable=too-many-arguments
async def pagination_handler(
    deserializer: typing.Callable[[typing.Any], typing.Any],
    direction: typing.Union[typing.Literal["before"], typing.Literal["after"]],
    request: typing.Callable[..., more_typing.Coroutine[typing.Any]],
    reversing: bool,
    start: str,
    maximum_limit: int,
    limit: typing.Optional[int] = None,
    id_getter: typing.Callable[[typing.Any], str] = lambda entity: str(entity.id),
) -> typing.AsyncIterator[typing.Any]:
    """Generate an async iterator for handling paginated endpoints.

    This will handle Discord's `before` and `after` pagination.

    Parameters
    ----------
    deserializer : typing.Callable[[typing.Any], typing.Any]
        The deserializer to use to deserialize raw elements.
    direction : typing.Union[`"before"`, `"after"`]
        The direction that this paginator should go in.
    request : typing.Callable[..., typing.Coroutine[typing.Any, typing.Any, typing.Any]]
        The `hikari.net.rest.REST` method that should be
        called to make requests for this paginator.
    reversing : bool
        Whether the retrieved array of objects should be reversed before
        iterating through it, this is needed for certain endpoints like
        `fetch_messages_before` where the order is static regardless of
        if you're using `before` or `after`.
    start : int
        The snowflake ID that this paginator should start at, `0` may be
        passed for `forward` pagination to start at the first created
        entity and `9223372036854775807` may be passed for `before` pagination
        to start at the newest entity (based on it's snowflake timestamp).
    maximum_limit : int
        The highest number that `limit` can be set to in a request for this
        specific endpoint.
    limit : int, optional
        The amount of deserialized entities that the iterator should return
        total, will be unlimited if set to `None`.
    id_getter : typing.Callable[[typing.Any], str]

    Returns
    -------
    typing.AsyncIterator[typing.Any]
        An async iterator of the found deserialized found objects.
    """
    while payloads := await request(
        limit=maximum_limit if limit is None or limit > maximum_limit else limit, **{direction: start},
    ):
        if reversing:
            payloads.reverse()
        if limit is not None:
            limit -= len(payloads)

        for payload in payloads:
            entity = deserializer(payload)
            yield entity
        if limit == 0:
            break
        start = id_getter(entity)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Basic transformation utilities."""

from __future__ import annotations

__all__ = [
    "nullable_cast",
    "try_cast",
    "try_cast_or_defer_unary_operator",
    "put_if_specified",
    "image_bytes_to_image_data",
    "parse_http_date",
    "parse_iso_8601_ts",
    "discord_epoch_to_datetime",
    "unix_epoch_to_datetime",
    "pluralize",
    "resolve_signature",
    "EMPTY",
]

import base64
import contextlib
import datetime
import email.utils
import functools
import inspect
import operator
import re
import typing

if typing.TYPE_CHECKING:
    import enum
    import types

    IntFlagT = typing.TypeVar("IntFlagT", bound=enum.IntFlag)
    RawIntFlagValueT = typing.Union[typing.AnyStr, typing.SupportsInt, int]
    CastInputT = typing.TypeVar("CastInputT")
    CastOutputT = typing.TypeVar("CastOutputT")
    DefaultT = typing.TypeVar("DefaultT")
    TypeCastT = typing.Callable[[CastInputT], CastOutputT]
    ResultT = typing.Union[CastOutputT, DefaultT]

DISCORD_EPOCH: typing.Final[int] = 1_420_070_400
ISO_8601_DATE_PART: typing.Final[typing.Pattern] = re.compile(r"^(\d{4})-(\d{2})-(\d{2})")
ISO_8601_TIME_PART: typing.Final[typing.Pattern] = re.compile(r"T(\d{2}):(\d{2}):(\d{2})(?:\.(\d{1,6}))?", re.I)
ISO_8601_TZ_PART: typing.Final[typing.Pattern] = re.compile(r"([+-])(\d{2}):(\d{2})$")


def nullable_cast(value: CastInputT, cast: TypeCastT, /) -> ResultT:
    """Attempt to cast the given `value` with the given `cast`.

    This will only succeed if `value` is not `None`. If it is `None`, then
    `None` is returned instead.
    """
    if value is None:
        return None
    return cast(value)


def try_cast(value: CastInputT, cast: TypeCastT, default: DefaultT = None, /) -> ResultT:
    """Try to cast the given value to the given cast.

    If it throws a `Exception` or derivative, it will return `default`
    instead of the cast value instead.
    """
    with contextlib.suppress(Exception):
        return cast(value)
    return default


def try_cast_or_defer_unary_operator(type_: typing.Type, /):
    """Return a unary operator that will try to cast the given input to the type provided.

    Parameters
    ----------
    type_ : typing.Callable[..., `output type`]
        The type to cast to.
    """
    return lambda data: try_cast(data, type_, data)


def put_if_specified(
    mapping: typing.Dict[typing.Hashable, typing.Any],
    key: typing.Hashable,
    value: typing.Any,
    type_after: typing.Optional[TypeCastT] = None,
    /,
) -> None:
    """Add a value to the mapping under the given key as long as the value is not `...`.

    Parameters
    ----------
    mapping : typing.Dict[typing.Hashable, typing.Any]
        The mapping to add to.
    key : typing.Hashable
        The key to add the value under.
    value : typing.Any
        The value to add.
    type_after : typing.Callable[[`input type`], `output type`], optional
        Type to apply to the value when added.
    """
    if value is not ...:
        if type_after:
            mapping[key] = type_after(value)
        else:
            mapping[key] = value


def image_bytes_to_image_data(img_bytes: typing.Optional[bytes] = None, /) -> typing.Optional[str]:
    """Encode image bytes into an image data string.

    Parameters
    ----------
    img_bytes : bytes, optional
        The image bytes.

    Raises
    ------
    ValueError
        If the image type passed is not supported.

    Returns
    -------
    str, optional
        The `image_bytes` given encoded into an image data string or
        `None`.

    !!! note
        Supported image types: `.png`, `.jpeg`, `.jfif`, `.gif`, `.webp`
    """
    if img_bytes is None:
        return None

    if img_bytes[:8] == b"\211PNG\r\n\032\n":
        img_type = "image/png"
    elif img_bytes[6:10] in (b"Exif", b"JFIF"):
        img_type = "image/jpeg"
    elif img_bytes[:6] in (b"GIF87a", b"GIF89a"):
        img_type = "image/gif"
    elif img_bytes.startswith(b"RIFF") and img_bytes[8:12] == b"WEBP":
        img_type = "image/webp"
    else:
        raise ValueError("Unsupported image type passed")

    image_data = base64.b64encode(img_bytes).decode()

    return f"data:{img_type};base64,{image_data}"


def parse_http_date(date_str: str, /) -> datetime.datetime:
    """Return the HTTP date as a datetime object.

    Parameters
    ----------
    date_str : str
        The RFC-2822 (section 3.3) compliant date string to parse.

    Returns
    -------
    datetime.datetime
        The HTTP date as a datetime object.

    References
    ----------
    [RFC 2822](https://www.ietf.org/rfc/rfc2822.txt)
    [Mozilla documentation for Date header](https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Date)
    """
    # According to Mozilla, these are always going to be GMT (which is UTC).
    return email.utils.parsedate_to_datetime(date_str).replace(tzinfo=datetime.timezone.utc)


def parse_iso_8601_ts(date_string: str, /) -> datetime.datetime:
    """Parse an ISO 8601 date string into a `datetime.datetime` object.

    Parameters
    ----------
    date_string : str
        The ISO 8601 compliant date string to parse.

    Returns
    -------
    datetime.datetime
        The ISO 8601 date string as a datetime object.

    References
    ----------
    [ISO 8601](https://en.wikipedia.org/wiki/ISO_8601)
    """
    year, month, day = map(int, ISO_8601_DATE_PART.findall(date_string)[0])

    time_part = ISO_8601_TIME_PART.findall(date_string)[0]
    hour, minute, second, partial = time_part

    # Pad the millisecond part if it is not in microseconds, otherwise Python will complain.
    partial = partial + (6 - len(partial)) * "0"
    hour, minute, second, partial = int(hour), int(minute), int(second), int(partial)
    if date_string.endswith(("Z", "z")):
        timezone = datetime.timezone.utc
    else:
        sign, tz_hour, tz_minute = ISO_8601_TZ_PART.findall(date_string)[0]
        tz_hour, tz_minute = int(tz_hour), int(tz_minute)
        offset = datetime.timedelta(hours=tz_hour, minutes=tz_minute)
        if sign == "-":
            offset = -offset
        timezone = datetime.timezone(offset)

    return datetime.datetime(year, month, day, hour, minute, second, partial, timezone)


def discord_epoch_to_datetime(epoch: int, /) -> datetime.datetime:
    """Parse a Discord epoch into a `datetime.datetime` object.

    Parameters
    ----------
    epoch : int
        Number of milliseconds since 1/1/2015 (UTC)

    Returns
    -------
    datetime.datetime
        Number of seconds since 1/1/1970 within a datetime object (UTC).
    """
    return datetime.datetime.fromtimestamp(epoch / 1000 + DISCORD_EPOCH, datetime.timezone.utc)


def unix_epoch_to_datetime(epoch: int, /) -> datetime.datetime:
    """Parse a UNIX epoch to a `datetime.datetime` object.

    Parameters
    ----------
    epoch : int
        Number of milliseconds since 1/1/1970 (UTC)

    Returns
    -------
    datetime.datetime
        Number of seconds since 1/1/1970 within a datetime object (UTC).
    """
    return datetime.datetime.fromtimestamp(epoch / 1000, datetime.timezone.utc)


def pluralize(count: int, name: str, suffix: str = "s") -> str:
    """Pluralizes a word."""
    return f"{count} {name + suffix}" if count - 1 else f"{count} {name}"


EMPTY: typing.Final[inspect.Parameter.empty] = inspect.Parameter.empty
"""A singleton that empty annotations will be set to in `resolve_signature`."""


def resolve_signature(func: typing.Callable) -> inspect.Signature:
    """Get the `inspect.Signature` of `func` with resolved forward annotations.

    Parameters
    ----------
    func : typing.Callable[[...], ...]
        The function to get the resolved annotations from.

    Returns
    -------
    typing.Signature
        A `typing.Signature` object with all forward reference annotations
        resolved.
    """
    signature = inspect.signature(func)
    resolved_type_hints = None
    parameters = []
    for key, value in signature.parameters.items():
        if isinstance(value.annotation, str):
            if resolved_type_hints is None:
                resolved_type_hints = typing.get_type_hints(func)
            parameters.append(value.replace(annotation=resolved_type_hints[key]))
        else:
            parameters.append(value)
    signature = signature.replace(parameters=parameters)

    if isinstance(signature.return_annotation, str):
        if resolved_type_hints is None:
            return_annotation = typing.get_type_hints(func)["return"]
        else:
            return_annotation = resolved_type_hints["return"]
        signature = signature.replace(return_annotation=return_annotation)

    return signature


def dereference_int_flag(
    int_flag_type: typing.Type[IntFlagT],
    raw_value: typing.Union[RawIntFlagValueT, typing.Collection[RawIntFlagValueT]],
) -> IntFlagT:
    """Cast to the provided `enum.IntFlag` type.

    This supports resolving bitfield integers as well as decoding a sequence
    of case insensitive flag names into one combined value.

    Parameters
    ----------
    int_flag_type : typing.Type[enum.IntFlag]
        The type of the int flag to check.
    raw_value : Castable Value
        The raw value to convert.

    Returns
    -------
    enum.IntFlag
        The cast value as a flag.

    !!! note
        Types that are a `Castable Value` include:
        - `str`
        - `int`
        - `typing.SupportsInt`
        - `typing.Collection`[`Castable Value`]

        When a collection is passed, values will be combined using functional
        reduction via the `operator.or_` operator.
    """
    if isinstance(raw_value, str) and raw_value.isdigit():
        raw_value = int(raw_value)

    if not isinstance(raw_value, int):
        raw_value = functools.reduce(operator.or_, (int_flag_type[name.upper()] for name in raw_value))

    return int_flag_type(raw_value)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Entities directly related to creating and managing gateway shards."""

from __future__ import annotations

__all__ = ["Activity", "GatewayBot", "SessionStartLimit"]

import datetime
import typing

import attr

from hikari import bases
from hikari import guilds
from hikari.internal import marshaller


def _rest_after_deserializer(after: int) -> datetime.timedelta:
    return datetime.timedelta(milliseconds=after)


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class SessionStartLimit(bases.HikariEntity, marshaller.Deserializable):
    """Used to represent information about the current session start limits."""

    total: int = marshaller.attrib(deserializer=int, repr=True)
    """The total number of session starts the current bot is allowed."""

    remaining: int = marshaller.attrib(deserializer=int, repr=True)
    """The remaining number of session starts this bot has."""

    reset_after: datetime.timedelta = marshaller.attrib(deserializer=_rest_after_deserializer, repr=True)
    """When `SessionStartLimit.remaining` will reset for the current bot.

    After it resets it will be set to `SessionStartLimit.total`.
    """


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GatewayBot(bases.HikariEntity, marshaller.Deserializable):
    """Used to represent gateway information for the connected bot."""

    url: str = marshaller.attrib(deserializer=str, repr=True)
    """The WSS URL that can be used for connecting to the gateway."""

    shard_count: int = marshaller.attrib(raw_name="shards", deserializer=int, repr=True)
    """The recommended number of shards to use when connecting to the gateway."""

    session_start_limit: SessionStartLimit = marshaller.attrib(
        deserializer=SessionStartLimit.deserialize, inherit_kwargs=True, repr=True
    )
    """Information about the bot's current session start limit."""


def _undefined_type_default() -> typing.Literal[guilds.ActivityType.PLAYING]:
    return guilds.ActivityType.PLAYING


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class Activity(marshaller.Deserializable, marshaller.Serializable):
    """An activity that the bot can set for one or more shards.

    This will show the activity as the bot's presence.
    """

    name: str = marshaller.attrib(deserializer=str, serializer=str, repr=True)
    """The activity name."""

    url: typing.Optional[str] = marshaller.attrib(
        deserializer=str, serializer=str, if_none=None, if_undefined=None, default=None, repr=True
    )
    """The activity URL. Only valid for `STREAMING` activities."""

    type: guilds.ActivityType = marshaller.attrib(
        deserializer=guilds.ActivityType,
        serializer=int,
        if_undefined=_undefined_type_default,
        default=guilds.ActivityType.PLAYING,
        repr=True,
    )
    """The activity type."""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Components and entities that are used to describe Discord gateway guild events."""

from __future__ import annotations

__all__ = [
    "GuildCreateEvent",
    "GuildUpdateEvent",
    "GuildLeaveEvent",
    "GuildUnavailableEvent",
    "BaseGuildBanEvent",
    "GuildBanAddEvent",
    "GuildBanRemoveEvent",
    "GuildEmojisUpdateEvent",
    "GuildIntegrationsUpdateEvent",
    "GuildMemberAddEvent",
    "GuildMemberUpdateEvent",
    "GuildMemberRemoveEvent",
    "GuildRoleCreateEvent",
    "GuildRoleUpdateEvent",
    "GuildRoleDeleteEvent",
    "PresenceUpdateEvent",
]

import abc
import typing

import attr

from hikari import bases as base_entities
from hikari import emojis as _emojis
from hikari import guilds
from hikari import intents
from hikari import unset
from hikari import users
from hikari.events import base as base_events
from hikari.internal import conversions
from hikari.internal import marshaller

if typing.TYPE_CHECKING:
    import datetime

    from hikari.internal import more_typing


@base_events.requires_intents(intents.Intent.GUILDS)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildCreateEvent(base_events.HikariEvent, guilds.Guild):
    """Used to represent Guild Create gateway events.

    Will be received when the bot joins a guild, and when a guild becomes
    available to a guild (either due to outage or at startup).
    """


@base_events.requires_intents(intents.Intent.GUILDS)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildUpdateEvent(base_events.HikariEvent, guilds.Guild):
    """Used to represent Guild Update gateway events."""


@base_events.requires_intents(intents.Intent.GUILDS)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildLeaveEvent(base_events.HikariEvent, base_entities.UniqueEntity, marshaller.Deserializable):
    """Fired when the current user leaves the guild or is kicked/banned from it.

    !!! note
        This is fired based on Discord's Guild Delete gateway event.
    """


@base_events.requires_intents(intents.Intent.GUILDS)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildUnavailableEvent(base_events.HikariEvent, base_entities.UniqueEntity, marshaller.Deserializable):
    """Fired when a guild becomes temporarily unavailable due to an outage.

    !!! note
        This is fired based on Discord's Guild Delete gateway event.
    """


@base_events.requires_intents(intents.Intent.GUILD_BANS)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class BaseGuildBanEvent(base_events.HikariEvent, marshaller.Deserializable, abc.ABC):
    """A base object that guild ban events will inherit from."""

    guild_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake, repr=True)
    """The ID of the guild this ban is in."""

    user: users.User = marshaller.attrib(deserializer=users.User.deserialize, inherit_kwargs=True, repr=True)
    """The object of the user this ban targets."""


@base_events.requires_intents(intents.Intent.GUILD_BANS)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildBanAddEvent(BaseGuildBanEvent):
    """Used to represent a Guild Ban Add gateway event."""


@base_events.requires_intents(intents.Intent.GUILD_BANS)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildBanRemoveEvent(BaseGuildBanEvent):
    """Used to represent a Guild Ban Remove gateway event."""


def _deserialize_emojis(
    payload: more_typing.JSONArray, **kwargs: typing.Any
) -> typing.Mapping[base_entities.Snowflake, _emojis.GuildEmoji]:
    return {base_entities.Snowflake(emoji["id"]): _emojis.GuildEmoji.deserialize(emoji, **kwargs) for emoji in payload}


@base_events.requires_intents(intents.Intent.GUILD_EMOJIS)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildEmojisUpdateEvent(base_events.HikariEvent, marshaller.Deserializable):
    """Represents a Guild Emoji Update gateway event."""

    guild_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake)
    """The ID of the guild this emoji was updated in."""

    emojis: typing.Mapping[base_entities.Snowflake, _emojis.GuildEmoji] = marshaller.attrib(
        deserializer=_deserialize_emojis, inherit_kwargs=True, repr=True
    )
    """The updated mapping of emojis by their ID."""


@base_events.requires_intents(intents.Intent.GUILD_INTEGRATIONS)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildIntegrationsUpdateEvent(base_events.HikariEvent, marshaller.Deserializable):
    """Used to represent Guild Integration Update gateway events."""

    guild_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake, repr=True)
    """The ID of the guild the integration was updated in."""


@base_events.requires_intents(intents.Intent.GUILD_MEMBERS)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildMemberAddEvent(base_events.HikariEvent, guilds.GuildMember):
    """Used to represent a Guild Member Add gateway event."""

    guild_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake, repr=True)
    """The ID of the guild where this member was added."""


def _deserialize_role_ids(payload: more_typing.JSONArray) -> typing.Sequence[base_entities.Snowflake]:
    return [base_entities.Snowflake(role_id) for role_id in payload]


@base_events.requires_intents(intents.Intent.GUILD_MEMBERS)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildMemberUpdateEvent(base_events.HikariEvent, marshaller.Deserializable):
    """Used to represent a Guild Member Update gateway event.

    Sent when a guild member or their inner user object is updated.
    """

    guild_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake, repr=True)
    """The ID of the guild this member was updated in."""

    role_ids: typing.Sequence[base_entities.Snowflake] = marshaller.attrib(
        raw_name="roles", deserializer=_deserialize_role_ids,
    )
    """A sequence of the IDs of the member's current roles."""

    user: users.User = marshaller.attrib(deserializer=users.User.deserialize, inherit_kwargs=True, repr=True)
    """The object of the user who was updated."""

    nickname: typing.Union[None, str, unset.Unset] = marshaller.attrib(
        raw_name="nick", deserializer=str, if_none=None, if_undefined=unset.Unset, default=unset.UNSET
    )
    """This member's nickname.

    When set to `None`, this has been removed and when set to
    `hikari.unset.UNSET` this hasn't been acted on.
    """

    premium_since: typing.Union[None, datetime.datetime, unset.Unset] = marshaller.attrib(
        deserializer=conversions.parse_iso_8601_ts, if_none=None, if_undefined=unset.Unset, default=unset.UNSET
    )
    """The datetime of when this member started "boosting" this guild.

    Will be `None` if they aren't boosting.
    """


@base_events.requires_intents(intents.Intent.GUILD_MEMBERS)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildMemberRemoveEvent(base_events.HikariEvent, marshaller.Deserializable):
    """Used to represent Guild Member Remove gateway events.

    Sent when a member is kicked, banned or leaves a guild.
    """

    # TODO: make GuildMember event into common base class.
    guild_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake, repr=True)
    """The ID of the guild this user was removed from."""

    user: users.User = marshaller.attrib(deserializer=users.User.deserialize, inherit_kwargs=True, repr=True)
    """The object of the user who was removed from this guild."""


@base_events.requires_intents(intents.Intent.GUILDS)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildRoleCreateEvent(base_events.HikariEvent, marshaller.Deserializable):
    """Used to represent a Guild Role Create gateway event."""

    guild_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake, repr=True)
    """The ID of the guild where this role was created."""

    role: guilds.GuildRole = marshaller.attrib(deserializer=guilds.GuildRole.deserialize, inherit_kwargs=True)
    """The object of the role that was created."""


@base_events.requires_intents(intents.Intent.GUILDS)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildRoleUpdateEvent(base_events.HikariEvent, marshaller.Deserializable):
    """Used to represent a Guild Role Create gateway event."""

    # TODO: make any event with a guild ID into a custom base event.
    # https://pypi.org/project/stupid/ could this work around the multiple inheritance problem?
    guild_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake, repr=True)
    """The ID of the guild where this role was updated."""

    role: guilds.GuildRole = marshaller.attrib(
        deserializer=guilds.GuildRole.deserialize, inherit_kwargs=True, repr=True
    )
    """The updated role object."""


@base_events.requires_intents(intents.Intent.GUILDS)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildRoleDeleteEvent(base_events.HikariEvent, marshaller.Deserializable):
    """Represents a gateway Guild Role Delete Event."""

    guild_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake, repr=True)
    """The ID of the guild where this role is being deleted."""

    role_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake, repr=True)
    """The ID of the role being deleted."""


@base_events.requires_intents(intents.Intent.GUILD_PRESENCES)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class PresenceUpdateEvent(base_events.HikariEvent, guilds.GuildMemberPresence):
    """Used to represent Presence Update gateway events.

    Sent when a guild member changes their presence.
    """



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Components used to make uploading data simpler.

What should I use?
------------------

- **"I have a file I want to read."**
    Use `FileStream`.
- **"The data is on a website or network resource."**
    Use `WebResourceStream`.
- **"The data is in an `io.BytesIO` or `io.StringIO`."**
    Use `ByteStream`.
- **"The data is a `bytes`, `bytearray`, `memoryview`, or `str`."**
    Use `ByteStream`.
- **"The data is provided in an async iterable or async iterator."**
    Use `ByteStream`.
- **"The data is in some other format."**
    Convert the data to one of the above descriptions, or implement your
    own provider by subclassing `BaseStream`.

How exactly do I use each one of these?
---------------------------------------
Check the documentation for each implementation to see examples and caveats.

Why is this so complicated?
---------------------------
Unfortunately, Python deals with async file IO in a really bad way. This means
that it is very easy to let IO operations impede the performance of your
application. Using these implementations correctly will enable you to mostly
offset that overhead.

General implications of not using these implementations can include increased
memory usage, and the application becoming unresponsive during IO.
"""

from __future__ import annotations

__all__ = ["BaseStream", "ByteStream", "WebResourceStream", "FileStream"]

import abc
import asyncio
import concurrent.futures
import functools
import http
import inspect
import io
import os
import typing

import aiohttp

from hikari import errors
from hikari.internal import more_asyncio

# XXX: find optimal size.
MAGIC_NUMBER: typing.Final[int] = 128 * 1024


class BaseStream(abc.ABC, typing.AsyncIterable[bytes]):
    """A data stream that can be uploaded in a message or downloaded.

    This is a wrapper for an async iterable of bytes.

    Parameters
    ----------
    filename : str
        The name of the stream.

    Implementations should provide an `__aiter__` method yielding chunks
    to upload. Chunks can be any size that is non-zero, but for performance
    should be around 64-256KiB in size.

    Example
    -------
        class HelloWorldStream(BaseStream):
            def __init__(self):
                super().__init__("hello-world.txt")

            async def __aiter__(self):
                for byte in b"hello, world!":
                    yield byte

        stream = HelloWorldStream()

    You can also use an implementation to read contents into memory

        >>> stream = HelloWorldStream()
        >>> data = await stream.read()
        >>> print(data)
        b"hello, world!"

    """

    filename: str
    """The name of the stream."""

    def __init__(self, filename: str) -> None:
        self.filename = filename

    def __repr__(self) -> str:
        return f"{type(self).__qualname__}(filename={self.filename!r})"

    async def read(self) -> bytes:
        """Return the entire contents of the data stream."""
        data = io.BytesIO()
        async for chunk in self:
            data.write(chunk)
        return data.getvalue()


class ByteStream(BaseStream):
    """A simple data stream that wraps something that gives bytes.

    For any asyncio-compatible stream that is an async iterator, or for
    any in-memory data, you can use this to safely upload the information
    in an API call.

    Parameters
    ----------
    filename : str
        The file name to use.
    obj : byte-like provider
        A provider of bytes-like objects. See the following sections for
        what is acceptable.

    A byte-like provider can be one of the following types:

    - A bytes-like object (see below)
    - **`AsyncIterator[bytes-like object]`**
    - **`AsyncIterable[bytes-like object]`**
    - **`AsyncGenerator[Any, bytes-like object]`** - an async generator that
        yields bytes-like objects.
    - **`io.BytesIO`**
    - **`io.StringIO`**

    A bytes-like object can be one of the following types:

    - **bytes**
    - **bytearray**
    - **str**
    - **memoryview**

    !!! warning
        Do not pass blocking IO streams!

        You should not use this to wrap any IO or file-like object other
        Standard Python file objects perform blocking IO, which will block
        the event loop each time a chunk is read.

        To read a file, use `FileStream` instead. This will read the file
        object incrementally, reducing memory usage significantly.

        Passing a different type of file object to this class may result in
        undefined behaviour.

    !!! note
        String objects get treated as UTF-8.

        String objects will always be treated as UTF-8 encoded byte objects.
        If you need to use a different encoding, you should transform the
        data manually into a bytes object first and pass the result to this
        class.

    !!! note
        Additional notes about performance.

        If you pass a bytes-like object, `io.BytesIO`, or `io.StringIO`, the
        resource will be transformed internally into a bytes object, and
        read in larger chunks using an `io.BytesIO` under the hood. This is done
        to increase performance, as yielding individual bytes would be very
        slow.

        If you pass an async iterator/iterable/generator directly, this will not
        be collected into chunks. Whatever is yielded will be the chunk that is
        uploaded. This allows for bit-inception with async iterators provided
        by other libraries tidily.

    Examples
    --------
    Passing bytes-like objects:

        >>> # A stream of bytes.
        >>> stream = ByteStream("hello.txt", b"hello, world!")

        >>> # A stream from a bytearray.
        >>> stream = ByteStream("hello.txt", bytearray(b"hello, world!"))

        >>> # A stream from a string. This will be treated as UTF-8 always.
        >>> stream = ByteStream("hello.txt", "hello, world!")

        >>> # A stream from an io.BytesIO
        >>> obj = io.BytesIO(some_data)
        >>> stream = ByteStream("cat.png", obj)

        >>> # A stream from an io.StringIO
        >>> obj = io.StringIO(some_data)
        >>> stream = ByteStream("some_text.txt", obj)

    Passing async iterators, iterables:

        >>> stream = ByteStream("cat.png", some_async_iterator)

        >>> stream = ByteStream("cat.png", some_async_iterable)

        >>> stream = ByteStream("cat.png", some_asyncio_stream_reader)

    Passing async generators:

        >>> async def asyncgen():
        ...    yield b"foo "
        ...    yield b"bar "

        >>> # You can pass the generator directly...
        >>> stream = ByteStream("foobar.txt", asyncgen())

        >>> # Or, if the generator function takes no parameters, you can pass the
        >>> # function reference instead.
        >>> stream = ByteStream("foobar.txt", asyncgen)

    Using a third-party non-blocking library such as `aiofiles` is possible
    if you can pass an async iterator:

        >>> async with aiofiles.open("cat.png", "rb") as afp:
        ...    stream = ByteStream("cat.png", afp)

    !!! warning
        Async iterators are read lazily. You should ensure in the latter
        example that the `afp` is not closed before you use the stream in a
        request.

    !!! note
        `aiofiles` is not included with this library, and serves as an example
        only. You can make use of `FileStream` instead if you need to read a
        file using non-blocking IO.

    """

    ___VALID_BYTE_TYPES___ = typing.Union[
        bytes, bytearray, str, memoryview,
    ]

    ___VALID_TYPES___ = typing.Union[
        typing.AsyncGenerator[typing.Any, ___VALID_BYTE_TYPES___],
        typing.AsyncIterator[___VALID_BYTE_TYPES___],
        typing.AsyncIterable[___VALID_BYTE_TYPES___],
        ___VALID_BYTE_TYPES___,
        io.BytesIO,
        io.StringIO,
    ]

    _obj: typing.Union[
        typing.AsyncGenerator[typing.Any, ___VALID_BYTE_TYPES___],
        typing.AsyncIterable[typing.Any, ___VALID_BYTE_TYPES___],
    ]

    def __init__(self, filename: str, obj: ___VALID_TYPES___) -> None:
        super().__init__(filename)

        if inspect.isasyncgenfunction(obj):
            self._obj = obj()
            return

        if more_asyncio.is_async_iterable(obj):
            obj = obj.__aiter__()

        if more_asyncio.is_async_iterator(obj):
            self._obj = self._aiter_async_iterator(obj)
            return

        if inspect.isasyncgen(obj):
            self._obj = obj
            return

        if isinstance(obj, (io.StringIO, io.BytesIO)):
            obj = obj.getvalue()

        if isinstance(obj, (str, memoryview, bytearray)):
            obj = self._to_bytes(obj)

        if isinstance(obj, bytes):
            self._obj = self._aiter_bytes(obj)
            return

        raise TypeError(f"Expected bytes-like object or async generator, got {type(obj).__qualname__}")

    async def __aiter__(self) -> typing.AsyncIterator[bytes]:
        async for chunk in self._obj:
            yield self._to_bytes(chunk)

    async def _aiter_async_iterator(
        self, async_iterator: typing.AsyncIterator[___VALID_BYTE_TYPES___]
    ) -> typing.AsyncIterator[bytes]:
        try:
            while True:
                yield self._to_bytes(await async_iterator.__anext__())
        except StopAsyncIteration:
            pass

    @staticmethod
    async def _aiter_bytes(bytes_: bytes) -> typing.AsyncIterator[bytes]:
        stream = io.BytesIO(bytes_)
        while chunk := stream.read(MAGIC_NUMBER):
            yield chunk

    @staticmethod
    def _to_bytes(byte_like: ___VALID_BYTE_TYPES___) -> bytes:
        if isinstance(byte_like, str):
            return bytes(byte_like, "utf-8")
        if isinstance(byte_like, memoryview):
            return byte_like.tobytes()
        if isinstance(byte_like, bytearray):
            return bytes(byte_like)
        if isinstance(byte_like, bytes):
            return byte_like
        raise TypeError(f"Expected bytes-like chunks, got {type(byte_like).__qualname__}")


class WebResourceStream(BaseStream):
    """An async iterable of bytes that is represented by a web resource.

    Using this to upload an attachment will lazily load chunks and send them
    using bit-inception, vastly reducing the memory overhead by not storing the
    entire resource in memory before sending it.

    Parameters
    ----------
    filename : str
        The file name to use.
    url : str
        The URL to the resource to stream.

    Example
    -------
        >>> stream = WebResourceStream("cat-not-found.png", "https://http.cat/404")
    """

    url: str
    """The URL of the resource."""

    def __init__(self, filename: str, url: str) -> None:
        super().__init__(filename)
        self.url = url

    async def __aiter__(self) -> typing.AsyncIterator[bytes]:
        async with aiohttp.request("GET", self.url) as response:
            if 200 <= response.status < 300:
                async for chunk in response.content:
                    yield chunk
                return

            raw_body = await response.read()

        real_url = str(response.real_url)

        if response.status == http.HTTPStatus.BAD_REQUEST:
            raise errors.BadRequest(real_url, response.headers, raw_body)
        if response.status == http.HTTPStatus.UNAUTHORIZED:
            raise errors.Unauthorized(real_url, response.headers, raw_body)
        if response.status == http.HTTPStatus.FORBIDDEN:
            raise errors.Forbidden(real_url, response.headers, raw_body)
        if response.status == http.HTTPStatus.NOT_FOUND:
            raise errors.NotFound(real_url, response.headers, raw_body)

        if 400 <= response.status < 500:
            cls = errors.ClientHTTPErrorResponse
        elif 500 <= response.status < 600:
            cls = errors.ServerHTTPErrorResponse
        else:
            cls = errors.HTTPErrorResponse

        raise cls(real_url, http.HTTPStatus(response.status), response.headers, raw_body)


class FileStream(BaseStream):
    r"""Asynchronous reader for a local file.

    Parameters
    ----------
    filename : str, optional
        The custom file name to give the file when uploading it. May be
        omitted.
    path : str OR os.PathLike
        The path-like object that describes the file to upload.

    executor : concurrent.futures.Executor, optional
        An optional executor to run the IO operations in. If not specified, the
        default executor for this loop will be used instead.

    Examples
    --------
    Providing an explicit custom filename:

        >>> # UNIX/Linux/MacOS users
        >>> FileStream("kitteh.png", "path/to/cat.png")
        >>> FileStream("kitteh.png", "/mnt/cat-pictures/cat.png")

        >>> # Windows users
        >>> FileStream("kitteh.png", r"Pictures\Cat.png")
        >>> FileStream("kitteh.png", r"C:\Users\CatPerson\Pictures\Cat.png")

    Inferring the filename from the file path:

        >>> # UNIX/Linux/MacOS users
        >>> FileStream("path/to/cat.png")
        >>> FileStream("/mnt/cat-pictures/cat.png")

        >>> # Windows users
        >>> FileStream(r"Pictures\Cat.png")
        >>> FileStream(r"C:\Users\CatPerson\Pictures\Cat.png")

    !!! note
        This implementation only provides the basis for READING
        a file without blocking. For writing to files asynchronously,
        you should consider using a third-party library such as
        `aiofiles` or `aiofile`, or using an `Executor` instead.

    !!! warning
        While it is possible to use a ProcessPoolExecutor executor
        implementation with this class, use is discouraged. Process pools
        can only communicate using pipes or the pickle protocol, which
        makes the vanilla Python implementation unsuitable for use with
        async iterators. Since file handles cannot be pickled, the use
        of a ProcessPoolExecutor will result in the entire file being read
        in one chunk, which increases memory usage drastically.
    """

    @typing.overload
    def __init__(
        self,
        filename: str,
        path: typing.Union[str, os.PathLike],
        *,
        executor: typing.Optional[concurrent.futures.Executor] = None,
    ) -> None:
        ...

    @typing.overload
    def __init__(
        self, path: typing.Union[str, os.PathLike], /, *, executor: typing.Optional[concurrent.futures.Executor] = None,
    ) -> None:
        ...

    def __init__(self, *args, executor=None) -> None:
        if len(args) == 1:
            super().__init__(os.path.basename(*args))
            self.path = args[0]
        else:
            name, path = args
            super().__init__(name)
            self.path = path
        self._executor = executor

    def __aiter__(self) -> typing.AsyncIterator[bytes]:
        loop = asyncio.get_event_loop()
        # We cant use a process pool in the same way we do a thread pool, as
        # we cannot pickle file objects that we pass between threads. This
        # method instead will stream the data via a pipe to us.
        if isinstance(self._executor, concurrent.futures.ProcessPoolExecutor):

            return self._processpool_strategy(loop)

        return self._threadpool_strategy(loop)

    async def _threadpool_strategy(self, loop):
        fp = await loop.run_in_executor(self._executor, functools.partial(open, self.path, "rb"))
        try:
            while chunk := await loop.run_in_executor(self._executor, fp.read, MAGIC_NUMBER):
                yield chunk
        finally:
            await loop.run_in_executor(self._executor, fp.close)

    async def _processpool_strategy(self, loop):
        yield await loop.run_in_executor(self._executor, self._read_all, self.path)

    @staticmethod
    def _read_all(path):
        with open(path, "rb") as fp:
            return fp.read()
